# Copyright 2018 Red Hat, Inc.
# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import os
import sys
import subprocess
import time
import configparser
import datetime
import dateutil.tz
import uuid
from unittest import mock

import fixtures
import jwt
import testtools
import sqlalchemy

from zuul.lib import encryption
from zuul.lib.keystorage import OIDCSigningKeys
from zuul.zk import ZooKeeperClient
from zuul.zk.locks import SessionAwareLock
from zuul.cmd.client import parse_cutoff

from tests.unit.test_launcher import LauncherBaseTestCase
from tests.base import BaseTestCase, ZuulTestCase
from tests.base import (
    FIXTURE_DIR,
    iterate_timeout,
    return_data,
    simple_layout,
)

from kazoo.exceptions import NoNodeError


class BaseClientTestCase(BaseTestCase):
    config_file = 'zuul.conf'
    config_with_zk = True

    def setUp(self):
        super(BaseClientTestCase, self).setUp()
        self.test_root = self.useFixture(fixtures.TempDir(
            rootdir=os.environ.get("ZUUL_TEST_ROOT"))).path
        self.config = configparser.ConfigParser()
        self.config.read(os.path.join(FIXTURE_DIR, self.config_file))
        if self.config_with_zk:
            self.config_add_zk()

    def config_add_zk(self):
        self.setupZK()
        self.config.add_section('zookeeper')
        self.config.set('zookeeper', 'hosts', self.zk_chroot_fixture.zk_hosts)
        self.config.set('zookeeper', 'session_timeout', '30')
        self.config.set('zookeeper', 'tls_cert',
                        self.zk_chroot_fixture.zookeeper_cert)
        self.config.set('zookeeper', 'tls_key',
                        self.zk_chroot_fixture.zookeeper_key)
        self.config.set('zookeeper', 'tls_ca',
                        self.zk_chroot_fixture.zookeeper_ca)


class TestTenantValidationClient(BaseClientTestCase):
    config_with_zk = True

    def test_client_tenant_conf_check(self):
        self.config.set(
            'scheduler', 'tenant_config',
            os.path.join(FIXTURE_DIR, 'config/tenant-parser/simple.yaml'))
        with open(os.path.join(self.test_root, 'tenant_ok.conf'), 'w') as f:
            self.config.write(f)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', os.path.join(self.test_root, 'tenant_ok.conf'),
             'tenant-conf-check'], stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.assertEqual(p.returncode, 0, 'The command must exit 0')

        self.config.set(
            'scheduler', 'tenant_config',
            os.path.join(FIXTURE_DIR, 'config/tenant-parser/invalid.yaml'))
        with open(os.path.join(self.test_root, 'tenant_ko.conf'), 'w') as f:
            self.config.write(f)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', os.path.join(self.test_root, 'tenant_ko.conf'),
             'tenant-conf-check'], stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.assertEqual(p.returncode, 1, "The command must exit 1")
        self.assertIn(
            b"expected a dictionary for dictionary", out,
            "Expected error message not found")

        self.config.set(
            'scheduler', 'tenant_config',
            os.path.join(FIXTURE_DIR, 'config/tenant-parser/semaphore.yaml'))
        with open(os.path.join(self.test_root, 'tenant_sem.conf'), 'w') as f:
            self.config.write(f)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', os.path.join(self.test_root, 'tenant_sem.conf'),
             'tenant-conf-check'], stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.assertEqual(p.returncode, 1, "The command must exit 1")
        self.assertIn(
            b"global semaphore", out,
            "Expected error message not found")


class TestWebTokenClient(BaseClientTestCase):
    config_file = 'zuul-admin-web.conf'

    def test_no_authenticator(self):
        """Test that token generation is not possible without authenticator"""
        old_conf = io.StringIO()
        self.config.write(old_conf)
        self.config.remove_section('auth zuul_operator')
        with open(os.path.join(self.test_root,
                               'no_zuul_operator.conf'), 'w') as f:
            self.config.write(f)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', os.path.join(self.test_root, 'no_zuul_operator.conf'),
             'create-auth-token',
             '--auth-config', 'zuul_operator',
             '--user', 'marshmallow_man',
             '--tenant', 'tenant_one', ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        old_conf.seek(0)
        self.config = configparser.ConfigParser()
        self.config.read_file(old_conf)
        self.assertEqual(p.returncode, 1, 'The command must exit 1')

    def test_unsupported_driver(self):
        """Test that token generation is not possible with wrong driver"""
        old_conf = io.StringIO()
        self.config.write(old_conf)
        self.config.add_section('auth someauth')
        self.config.set('auth someauth', 'driver', 'RS256withJWKS')
        with open(os.path.join(self.test_root, 'JWKS.conf'), 'w') as f:
            self.config.write(f)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', os.path.join(self.test_root, 'JWKS.conf'),
             'create-auth-token',
             '--auth-config', 'someauth',
             '--user', 'marshmallow_man',
             '--tenant', 'tenant_one', ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        old_conf.seek(0)
        self.config = configparser.ConfigParser()
        self.config.read_file(old_conf)
        self.assertEqual(p.returncode, 1, 'The command must exit 1')

    def test_token_generation(self):
        """Test token generation"""
        with open(os.path.join(self.test_root, 'good.conf'), 'w') as f:
            self.config.write(f)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', os.path.join(self.test_root, 'good.conf'),
             'create-auth-token',
             '--auth-conf', 'zuul_operator',
             '--user', 'marshmallow_man',
             '--tenant', 'tenant_one',
             '--claim', 'claim1:value1',
             '--claim', 'claim2:value2',
             '--print-meta-info', ],
            stdout=subprocess.PIPE)
        now = time.time()
        out, _ = p.communicate()
        self.assertEqual(p.returncode, 0, 'The command must exit 0')
        self.assertTrue(out.startswith(b"Bearer "), out)
        # Get the token from the first line of the output
        token = jwt.decode(out.split(b'\n')[0][len("Bearer "):],
                           key=self.config.get(
                               'auth zuul_operator',
                               'secret'),
                           algorithms=[self.config.get(
                               'auth zuul_operator',
                               'driver')],
                           audience=self.config.get(
                               'auth zuul_operator',
                               'client_id'),)
        self.assertEqual('marshmallow_man', token.get('sub'))
        self.assertEqual('zuul_operator', token.get('iss'))
        self.assertEqual('zuul.example.com', token.get('aud'))
        self.assertEqual('value1', token.get('claim1'))
        self.assertEqual('value2', token.get('claim2'))
        admin_tenants = token.get('zuul', {}).get('admin', [])
        self.assertTrue('tenant_one' in admin_tenants, admin_tenants)
        # allow one minute for the process to run
        self.assertTrue(580 <= int(token['exp']) - now < 660,
                        (token['exp'], now))

        # Get the meta info from the second line of the output
        meta_info_header = out.split(b'\n')[1]
        self.assertIn("Meta Info", meta_info_header.decode('utf-8'))

    def test_token_generation_no_zuul_admin(self):
        """Test token generation without zuul.admin claim"""
        with open(os.path.join(self.test_root, 'good.conf'), 'w') as f:
            self.config.write(f)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', os.path.join(self.test_root, 'good.conf'),
             'create-auth-token',
             '--auth-conf', 'zuul_operator',
             '--user', 'marshmallow_man',],
            stdout=subprocess.PIPE,)
        now = time.time()
        out, _ = p.communicate()
        self.assertEqual(p.returncode, 0, 'The command must exit 0')
        self.assertTrue(out.startswith(b"Bearer "), out)
        # Get the token from the first line of the output
        token = jwt.decode(out.split(b'\n')[0][len("Bearer "):],
                           key=self.config.get(
                               'auth zuul_operator',
                               'secret'),
                           algorithms=[self.config.get(
                               'auth zuul_operator',
                               'driver')],
                           audience=self.config.get(
                               'auth zuul_operator',
                               'client_id'),)
        self.assertEqual('marshmallow_man', token.get('sub'))
        self.assertEqual('zuul_operator', token.get('iss'))
        self.assertEqual('zuul.example.com', token.get('aud'))
        self.assertNotIn('zuul', token)
        # allow one minute for the process to run
        self.assertTrue(580 <= int(token['exp']) - now < 660,
                        (token['exp'], now))


class TestKeyOperations(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_export_import(self):
        # Test a round trip export/import of keys
        export_root = os.path.join(self.test_root, 'export')
        config_file = os.path.join(self.test_root, 'zuul.conf')
        with open(config_file, 'w') as f:
            self.config.write(f)

        # Save a copy of the keys in ZK
        old_data = self.getZKTree('/keystorage')

        # Export keys
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'export-keys', export_root],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        # Delete keys from ZK
        self.zk_client.client.delete('/keystorage', recursive=True)

        # Make sure it's really gone
        with testtools.ExpectedException(NoNodeError):
            self.getZKTree('/keystorage')

        # Import keys
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'import-keys', export_root],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        # Make sure the new data matches the original
        new_data = self.getZKTree('/keystorage')
        self.assertEqual(new_data, old_data)

    def test_copy_delete(self):
        config_file = os.path.join(self.test_root, 'zuul.conf')
        with open(config_file, 'w') as f:
            self.config.write(f)

        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'copy-keys',
             'gerrit', 'org/project',
             'gerrit', 'neworg/newproject',
             ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        data = self.getZKTree('/keystorage')
        self.assertEqual(
            data['/keystorage/gerrit/org/org%2Fproject/secrets'],
            data['/keystorage/gerrit/neworg/neworg%2Fnewproject/secrets'])
        self.assertEqual(
            data['/keystorage/gerrit/org/org%2Fproject/ssh'],
            data['/keystorage/gerrit/neworg/neworg%2Fnewproject/ssh'])

        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'delete-keys',
             'gerrit', 'org/project',
             ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        data = self.getZKTree('/keystorage')
        self.assertIsNone(
            data.get('/keystorage/gerrit/org/org%2Fproject/secrets'))
        self.assertIsNone(
            data.get('/keystorage/gerrit/org/org%2Fproject/ssh'))
        self.assertIsNone(
            data.get('/keystorage/gerrit/org/org%2Fproject'))
        # Ensure that deleting one project in a tree doesn't remove other
        # projects in that tree.
        self.assertIsNotNone(
            data.get('/keystorage/gerrit/org/org%2Fproject1'))
        self.assertIsNotNone(
            data.get('/keystorage/gerrit/org/org%2Fproject2'))
        self.assertIsNotNone(
            data.get('/keystorage/gerrit/org'))

        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'delete-keys',
             'gerrit', 'org/project1',
             ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'delete-keys',
             'gerrit', 'org/project2',
             ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        data = self.getZKTree('/keystorage')
        # Ensure that the last project being removed also removes its
        # org prefix entry.
        self.assertIsNone(
            data.get('/keystorage/gerrit/org/org%2Fproject1'))
        self.assertIsNone(
            data.get('/keystorage/gerrit/org/org%2Fproject2'))
        self.assertIsNone(
            data.get('/keystorage/gerrit/org'))


class TestOIDCSigningKeyOperation(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_delete_oidc_signing_keys(self):
        algorithm = "RS256"
        keystore = self.scheds.first.sched.keystore

        # Makesure keys are created by calling getLatestOidcSigningKeys()
        private_key1, _, version1 = keystore.getLatestOidcSigningKeys(
            algorithm)

        with keystore.createZKContext() as context:
            # Check that the keys are there
            test_keys1 = OIDCSigningKeys.loadKeys(
                context, algorithm)
            self.assertEqual(len(test_keys1.keys), 1)
            self.assertEqual(version1, 0)

            # Config and exeute the delete command
            config_file = os.path.join(self.test_root, 'zuul.conf')
            with open(config_file, 'w') as f:
                self.config.write(f)

            p = subprocess.Popen(
                [os.path.join(sys.prefix, 'bin/zuul-admin'),
                 '-c', config_file,
                 'delete-oidc-signing-keys',
                 algorithm,
                 ],
                stdout=subprocess.PIPE)
            out, _ = p.communicate()
            self.log.debug(out.decode('utf8'))
            self.assertEqual(p.returncode, 0)

            # Keys should be gone
            test_keys2 = OIDCSigningKeys.loadKeys(
                context, algorithm)
            self.assertIsNone(test_keys2)

        # New keys are auto generated when calling getLatestOidcSigningKeys()
        private_key2, _, version2 = keystore.getLatestOidcSigningKeys(
            algorithm)
        self.assertEqual(4096, private_key2.key_size)
        self.assertNotEqual(
            encryption.serialize_rsa_private_key(private_key2),
            encryption.serialize_rsa_private_key(private_key1))
        self.assertEqual(version2, 0)


class TestOfflineZKOperations(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def shutdown(self):
        pass

    def assertFinalState(self):
        pass

    def assertCleanShutdown(self):
        pass

    def test_delete_state(self):
        # Shut everything down (as much as possible) to reduce
        # logspam and errors.
        ZuulTestCase.shutdown(self)

        # Re-start the client connection because we need one for the
        # test.
        self.zk_client = ZooKeeperClient.fromConfig(self.config)
        self.zk_client.connect()

        config_file = os.path.join(self.test_root, 'zuul.conf')
        with open(config_file, 'w') as f:
            self.config.write(f)

        # Save a copy of the keys in ZK
        old_data = self.getZKTree('/keystorage')

        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'delete-state',
             ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE)
        out, _ = p.communicate(b'yes\n')
        self.log.debug(out.decode('utf8'))

        # Make sure the keys are still around
        new_data = self.getZKTree('/keystorage')
        self.assertEqual(new_data, old_data)

        # Make sure we really deleted everything
        with testtools.ExpectedException(NoNodeError):
            self.getZKTree('/zuul')

        self.zk_client.disconnect()

    def test_delete_state_keep_config_cache(self):
        # Shut everything down (as much as possible) to reduce
        # logspam and errors.
        ZuulTestCase.shutdown(self)

        # Re-start the client connection because we need one for the
        # test.
        self.zk_client = ZooKeeperClient.fromConfig(self.config)
        self.zk_client.connect()

        config_file = os.path.join(self.test_root, 'zuul.conf')
        with open(config_file, 'w') as f:
            self.config.write(f)

        # Save a copy of the things we keep in ZK
        old_keys = self.getZKTree('/keystorage')
        old_config = self.getZKTree('/zuul/config')

        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'delete-state', '--keep-config-cache',
             ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE)
        out, _ = p.communicate(b'yes\n')
        self.log.debug(out.decode('utf8'))

        # Make sure the keys are still around
        new_keys = self.getZKTree('/keystorage')
        self.assertEqual(new_keys, old_keys)
        new_config = self.getZKTree('/zuul/config')
        self.assertEqual(new_config, old_config)

        # Make sure we really deleted everything
        children = self.zk_client.client.get_children('/zuul')
        self.assertEqual(['config'], children)

        self.zk_client.disconnect()


class TestOnlineZKOperations(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def assertSQLState(self):
        pass

    def _test_delete_pipeline(self, pipeline):
        sched = self.scheds.first.sched
        tenant = sched.abide.tenants['tenant-one']
        # Force a reconfiguration due to a config change (so that the
        # tenant trigger event queue gets a minimum timestamp set)
        file_dict = {'zuul.yaml': ''}
        M = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        M.setMerged()
        self.fake_gerrit.addEvent(M.getChangeMergedEvent())
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        if pipeline == 'check':
            self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        else:
            A.addApproval('Code-Review', 2)
            self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Lock the check pipeline so we don't process the event we're
        # about to submit (it should go into the pipeline trigger event
        # queue and stay there while we delete the pipeline state).
        # This way we verify that events arrived before the deletion
        # still work.
        plock = SessionAwareLock(
            self.zk_client.client,
            f"/zuul/locks/pipeline/{tenant.name}/{pipeline}")
        plock.acquire(blocking=True, timeout=None)
        try:
            self.log.debug('Got pipeline lock')
            # Add a new event while our old last reconfigure time is
            # in place.
            B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
            if pipeline == 'check':
                self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
            else:
                B.addApproval('Code-Review', 2)
                self.fake_gerrit.addEvent(B.addApproval('Approved', 1))

            # Wait until it appears in the pipeline trigger event queue
            self.log.debug('Waiting for event')
            for x in iterate_timeout(30, 'trigger event queue has events'):
                if sched.pipeline_trigger_events[
                        tenant.name][pipeline].hasEvents():
                    break
            self.log.debug('Got event')
        except Exception:
            plock.release()
            raise
        # Grab the run handler lock so that we will continue to avoid
        # further processing of the event after we release the
        # pipeline lock (which the delete command needs to acquire).
        sched.run_handler_lock.acquire()
        try:
            plock.release()
            self.log.debug('Got run lock')
            config_file = os.path.join(self.test_root, 'zuul.conf')
            with open(config_file, 'w') as f:
                self.config.write(f)

            # Make sure the pipeline exists
            self.getZKTree(
                f'/zuul/tenant/{tenant.name}/pipeline/{pipeline}/item')
            # Save the old layout uuid
            tenant = sched.abide.tenants[tenant.name]
            old_layout_uuid = tenant.layout.uuid
            self.log.debug('Deleting pipeline state')
            p = subprocess.Popen(
                [os.path.join(sys.prefix, 'bin/zuul-admin'),
                 '-c', config_file,
                 'delete-pipeline-state',
                 tenant.name, pipeline,
                 ],
                stdout=subprocess.PIPE)
            # Delete the pipeline state
            out, _ = p.communicate()
            self.log.debug(out.decode('utf8'))
            self.assertEqual(p.returncode, 0, 'The command must exit 0')
            # Make sure it's deleted
            with testtools.ExpectedException(NoNodeError):
                self.getZKTree(
                    f'/zuul/tenant/{tenant.name}/pipeline/{pipeline}/item')
            # Make sure the change list is re-created
            self.getZKTree(
                f'/zuul/tenant/{tenant.name}/pipeline/{pipeline}/change_list')
        finally:
            sched.run_handler_lock.release()

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-merge', result='SUCCESS', changes='2,1'),
            dict(name='project-merge', result='SUCCESS', changes='3,1'),
            dict(name='project-test1', result='SUCCESS', changes='3,1'),
            dict(name='project-test2', result='SUCCESS', changes='3,1'),
        ], ordered=False)
        tenant = sched.abide.tenants[tenant.name]
        new_layout_uuid = tenant.layout.uuid
        self.assertEqual(old_layout_uuid, new_layout_uuid)
        self.assertEqual(
            tenant.layout.pipeline_managers[pipeline].state.layout_uuid,
            old_layout_uuid)

    def test_delete_pipeline_check(self):
        self._test_delete_pipeline('check')

    def test_delete_pipeline_gate(self):
        self._test_delete_pipeline('gate')


class TestDBPruneParse(BaseTestCase):
    def test_db_prune_parse(self):
        now = datetime.datetime(year=2023, month=5, day=28,
                                hour=22, minute=15, second=1,
                                tzinfo=dateutil.tz.tzutc())
        reference = datetime.datetime(year=2022, month=5, day=28,
                                      hour=22, minute=15, second=1,
                                      tzinfo=dateutil.tz.tzutc())
        # Test absolute times
        self.assertEqual(
            reference,
            parse_cutoff(now, '2022-05-28 22:15:01 UTC', None))
        self.assertEqual(
            reference,
            parse_cutoff(now, '2022-05-28 22:15:01', None))

        # Test relative times
        self.assertEqual(reference,
                         parse_cutoff(now, None, '8760h'))
        self.assertEqual(reference,
                         parse_cutoff(now, None, '365d'))
        with testtools.ExpectedException(RuntimeError):
            self.assertEqual(reference,
                             parse_cutoff(now, None, '1y'))


class DBPruneTestCase(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'
    # This should be larger than the limit size in sqlconnection
    num_buildsets = 55

    def _createBuildset(self, update_time):
        connection = self.scheds.first.sched.sql.connection
        buildset_uuid = uuid.uuid4().hex
        event_id = uuid.uuid4().hex
        tenant = 'tenant-one',
        event_type = "reconfiguration"
        description = "A reconfiguration happened"

        with connection.getSession() as db:
            start_time = update_time - datetime.timedelta(seconds=1)
            end_time = update_time
            db_buildset = db.createBuildSet(
                uuid=buildset_uuid,
                tenant=tenant,
                pipeline='check',
                event_id=event_id,
                event_timestamp=update_time,
                updated=update_time,
                first_build_start_time=start_time,
                last_build_end_time=end_time,
                result='SUCCESS',
            )
            db_ref = db.getOrCreateRef(
                project='org/project',
                ref='refs/changes/1',
                change=1,
                patchset='1',
                oldrev='',
                newrev='',
                branch='master',
                ref_url='http://gerrit.example.com/1',
            )
            db_buildset.refs.append(db_ref)

            for build_num in range(2):
                build_uuid = uuid.uuid4().hex
                db_build = db_buildset.createBuild(
                    ref=db_ref,
                    uuid=build_uuid,
                    job_name=f'job{build_num}',
                    start_time=start_time,
                    end_time=end_time,
                    result='SUCCESS',
                    voting=True,
                )
                for art_num in range(2):
                    db_build.createArtifact(
                        name=f'artifact{art_num}',
                        url='http://example.com',
                    )
                for provides_num in range(2):
                    db_build.createProvides(
                        name=f'item{provides_num}',
                    )
                for event_num in range(2):
                    db_build.createBuildEvent(
                        event_type=f'event{event_num}',
                        event_time=start_time,
                    )
            db.createSystemEvent(
                tenant=tenant,
                event_id=event_id,
                event_time=update_time,
                event_type=event_type,
                description=description,
            )

    def _query(self, db, model):
        table = model.__table__
        q = db.session().query(model).order_by(table.c.id.desc())
        try:
            return q.all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def _getBuildsets(self, db):
        return self._query(db, db.connection.buildSetModel)

    def _getBuilds(self, db):
        return self._query(db, db.connection.buildModel)

    def _getProvides(self, db):
        return self._query(db, db.connection.providesModel)

    def _getArtifacts(self, db):
        return self._query(db, db.connection.artifactModel)

    def _getBuildEvents(self, db):
        return self._query(db, db.connection.buildEventModel)

    def _getSystemEvents(self, db):
        return self._query(db, db.connection.systemEventModel)

    def _setup(self):
        config_file = os.path.join(self.test_root, 'zuul.conf')
        with open(config_file, 'w') as f:
            self.config.write(f)

        update_time = (datetime.datetime.utcnow() -
                       datetime.timedelta(minutes=self.num_buildsets))
        for x in range(self.num_buildsets):
            update_time = update_time + datetime.timedelta(minutes=1)
            self._createBuildset(update_time)

        connection = self.scheds.first.sched.sql.connection
        with connection.getSession() as db:
            buildsets = self._getBuildsets(db)
            builds = self._getBuilds(db)
            artifacts = self._getArtifacts(db)
            provides = self._getProvides(db)
            events = self._getBuildEvents(db)
            system_events = self._getSystemEvents(db)
        self.assertEqual(len(buildsets), self.num_buildsets)
        self.assertEqual(len(builds), 2 * self.num_buildsets)
        self.assertEqual(len(artifacts), 4 * self.num_buildsets)
        self.assertEqual(len(provides), 4 * self.num_buildsets)
        self.assertEqual(len(events), 4 * self.num_buildsets)
        self.assertEqual(len(system_events), self.num_buildsets)
        for build in builds:
            self.log.debug("Build %s %s %s",
                           build, build.start_time, build.end_time)
        return config_file

    def test_db_prune_before(self):
        # Test pruning buildsets before a specific date
        config_file = self._setup()
        connection = self.scheds.first.sched.sql.connection

        # Builds are reverse ordered; 0 is most recent
        buildsets = connection.getBuildsets()
        start_time = buildsets[0].first_build_start_time
        self.log.debug("Cutoff %s", start_time)

        # Use the default batch size (omit --batch-size arg)
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'prune-database',
             '--before', str(start_time),
             ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))

        with connection.getSession() as db:
            buildsets = self._getBuildsets(db)
            builds = self._getBuilds(db)
            artifacts = self._getArtifacts(db)
            provides = self._getProvides(db)
            events = self._getBuildEvents(db)
            system_events = self._getSystemEvents(db)
        for build in builds:
            self.log.debug("Build %s %s %s",
                           build, build.start_time, build.end_time)
        self.assertEqual(len(buildsets), 1)
        self.assertEqual(len(builds), 2)
        self.assertEqual(len(artifacts), 4)
        self.assertEqual(len(provides), 4)
        self.assertEqual(len(events), 4)
        self.assertEqual(len(system_events), 1)

    def test_db_prune_older_than(self):
        # Test pruning buildsets older than a relative time
        config_file = self._setup()
        connection = self.scheds.first.sched.sql.connection

        # We use 0d as the relative time here since the earliest we
        # support is 1d and that's tricky in unit tests.  The
        # prune_before test handles verifying that we don't just
        # always delete everything.
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'prune-database',
             '--older-than', '0d',
             '--batch-size', '5',
             ],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))

        with connection.getSession() as db:
            buildsets = self._getBuildsets(db)
            builds = self._getBuilds(db)
            artifacts = self._getArtifacts(db)
            provides = self._getProvides(db)
            events = self._getBuildEvents(db)
            system_events = self._getSystemEvents(db)
        self.assertEqual(len(buildsets), 0)
        self.assertEqual(len(builds), 0)
        self.assertEqual(len(artifacts), 0)
        self.assertEqual(len(provides), 0)
        self.assertEqual(len(events), 0)
        self.assertEqual(len(system_events), 0)


class TestDBPruneMysql(DBPruneTestCase):
    config_file = 'zuul-sql-driver-mysql.conf'


class TestDBPrunePostgres(DBPruneTestCase):
    config_file = 'zuul-sql-driver-postgres.conf'


class TestImageOperations(LauncherBaseTestCase):
    config_file = 'zuul-connections-nodepool.conf'
    tenant_config_file = 'config/multi-tenant-provider/main.yaml'

    @simple_layout('layouts/nodepool-image.yaml', enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        LauncherBaseTestCase.debian_return_data,
    )
    @return_data(
        'build-ubuntu-local-image',
        'refs/heads/master',
        LauncherBaseTestCase.ubuntu_return_data,
    )
    @mock.patch('zuul.driver.aws.awsendpoint.AwsImageUploadJob.run',
                return_value="test_external_id")
    def test_image_export_import(self, mock_image_upload_run):
        # Test a round trip export/import of images
        self.addImageBuildEvent(
            'tenant-one',
            'review.example.com/org/common-config',
            'master',
            ['debian-local', 'ubuntu-local'],
        )
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='build-debian-local-image', result='SUCCESS'),
            dict(name='build-ubuntu-local-image', result='SUCCESS'),
        ], ordered=False)

        export_root = os.path.join(self.test_root, 'export')
        config_file = os.path.join(self.test_root, 'zuul.conf')
        with open(config_file, 'w') as f:
            self.config.write(f)

        # Save a copy of ZK (only the data, not the locks dirs)
        old_images = self.getZKTree('/zuul/images/artifacts')
        old_uploads = self.getZKTree('/zuul/image-uploads/uploads')

        # Stop the launcher so we don't touch ZK
        self.launcher.stop()
        self.launcher.join()

        # Export images
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'export-images', export_root],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        # Delete images from ZK
        self.zk_client.client.delete('/zuul/images', recursive=True)
        self.zk_client.client.delete('/zuul/image-uploads', recursive=True)

        # Make sure it's really gone
        with testtools.ExpectedException(NoNodeError):
            self.getZKTree('/zuul/images')
        with testtools.ExpectedException(NoNodeError):
            self.getZKTree('/zuul/image-uploads')

        # Import keys
        p = subprocess.Popen(
            [os.path.join(sys.prefix, 'bin/zuul-admin'),
             '-c', config_file,
             'import-images', export_root],
            stdout=subprocess.PIPE)
        out, _ = p.communicate()
        self.log.debug(out.decode('utf8'))
        self.assertEqual(p.returncode, 0)

        # Make sure the new data matches the original
        new_images = self.getZKTree('/zuul/images/artifacts')
        self.assertEqual(old_images, new_images)
        new_uploads = self.getZKTree('/zuul/image-uploads/uploads')
        self.assertEqual(old_uploads, new_uploads)

        # Restart the launcher for graceful test shutdown
        self.launcher = self.createLauncher()
        self.waitUntilSettled()



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import itertools
import json
import logging
import time
from contextlib import suppress

import cachetools
import kazoo
import paramiko

from zuul.exceptions import AlgorithmNotSupportedException
from zuul.lib import encryption, strings
from zuul.zk import ZooKeeperBase
from zuul.zk.cache import ZuulTreeCache
from zuul.zk.zkobject import ZKContext, ZKObject

RSA_KEY_SIZE = 2048


class OIDCSigningKeysCache(ZuulTreeCache):

    def __init__(self, client):
        self.cache_updated_listeners = []
        super().__init__(
            client, OIDCSigningKeys.OIDC_ROOT_PATH, async_worker=False)

    def addCacheUpdatedListener(self, listener):
        self.cache_updated_listeners.append(listener)

    def postCacheHook(self, event, data, stat, key, obj):
        for listener in self.cache_updated_listeners:
            listener.onCacheUpdated()

    def objectFromRaw(self, key, data, zstat):
        return OIDCSigningKeys._fromRaw(self._zk_context, data, zstat, None)

    def updateFromRaw(self, obj, key, data, zstat):
        obj._updateFromRaw(self._zk_context, data, zstat, None)

    def parsePath(self, path):
        return self._formatKey(path.split('/')[-1]), True

    def getSigningKeys(self, algorithm):
        return self._cached_objects.get(self._formatKey(algorithm))

    def _formatKey(self, algorithm):
        # key format: ("oidc_keys", algorithm)
        return ('oidc_keys', algorithm)


class OIDCSigningKeys(ZKObject):

    OIDC_ROOT_PATH = "/keystorage-oidc"

    def __init__(self):
        super().__init__()

        self._set(
            schema=1,
            keys=[],
            algorithm=None,
        )

    def getPath(self):
        return self._getPathByAlgorithm(self.algorithm)

    def serialize(self, context):
        keydata = {
            'schema': self.schema,
            'keys': self.keys,
        }
        return json.dumps(keydata, sort_keys=True).encode("utf8")

    def appendKey(self, private_key, version):
        if not self._active_context:
            raise Exception("appendKey must be used with a context manager")
        key_dict = self._generateKeyDict(private_key, version)
        self.keys.append(key_dict)

    @classmethod
    def new(klass, context, algorithm, private_key, version=0):
        obj = klass()
        obj._set(algorithm=algorithm)
        obj._set(keys=[klass._generateKeyDict(private_key, version)])
        data = obj._trySerialize(context)
        obj._save(context, data, create=True)
        return obj

    @classmethod
    def loadKeys(klass, context, algorithm):
        path = klass._getPathByAlgorithm(algorithm)
        try:
            return klass.fromZK(context, path, algorithm=algorithm)
        except kazoo.exceptions.NoNodeError:
            return None

    @classmethod
    def createAndStoreKeys(
        klass, context, algorithm, password_bytes):
        context.log.debug(
            "Creating OIDC signing keys for algorithm: %s", algorithm)

        private_key = klass._generatePrivateKey(algorithm, password_bytes)
        klass.new(context, algorithm, private_key)

    @classmethod
    def rotateKeys(
        klass, context, algorithm, password_bytes, rotation_interval, max_ttl):
        context.log.debug("Rotating OIDC signing keys, algorithm: %s,"
                          "rotation_interval: %s, max_ttl: %s",
                          algorithm, rotation_interval, max_ttl)
        key_data = klass.loadKeys(context, algorithm)

        if not key_data:
            klass.createAndStoreKeys(context, algorithm, password_bytes)
        else:
            # Check if there are old keys needs to be deleted
            # Here also need to handle the corner case when max_ttl
            # is bigger than rotation_interval. In this case, multiple
            # valid keys can exist at the same time. In either case,
            # find the last key whose created time is smaller than the
            # current time - max_ttl, then all tokens signed by the
            # keys before it should be expired and we can remove them.
            older_than = int(time.time()) - max_ttl

            # Make sure keys are sorted by created time
            existing_keys = sorted(
                key_data.keys, key=lambda key: key['created'])

            with key_data.activeContext(context):
                for index in range(len(existing_keys) - 1, -1, -1):
                    key = existing_keys[index]
                    if key["created"] < older_than and index > 0:
                        context.log.info(
                            "Removing OIDC keys whose version is older"
                            "than %s", existing_keys[index]["version"])
                        key_data._set(keys=existing_keys[index:])
                        break

                # Check if latest key is outdated and create a new one
                latest_key = key_data.keys[-1]
                age_seconds = int(time.time()) - latest_key["created"]
                if age_seconds > rotation_interval:
                    new_version = latest_key["version"] + 1
                    context.log.info(
                        "Generating new OIDC key with version %s", new_version)
                    private_key = OIDCSigningKeys._generatePrivateKey(
                        algorithm, password_bytes)
                    key_data.appendKey(
                        private_key=private_key,
                        version=new_version)

    @classmethod
    def deleteKeysByAlgorithm(klass, context, algorithm):
        path = klass._getPathByAlgorithm(algorithm)
        klass._delete(context, path)

    @classmethod
    def _getPathByAlgorithm(klass, algorithm):
        return f"{klass.OIDC_ROOT_PATH}/{algorithm}"

    @staticmethod
    def _generateKeyDict(private_key, version):
        return {
            "version": version,
            "created": int(time.time()),
            "private_key": private_key,
        }

    @classmethod
    def _generatePrivateKey(klass, algorithm, password_bytes):
        if algorithm == "RS256":
            private_key, public_key = encryption.generate_rsa_keypair()
            pem_private_key = encryption.serialize_rsa_private_key(
                private_key, password_bytes)
            return pem_private_key.decode("utf-8")
        else:
            raise AlgorithmNotSupportedException(
                f"Algorithm {algorithm} is not supported")

    def __eq__(self, other):
        return (isinstance(other, OIDCSigningKeys) and
                self.schema == other.schema and
                self.keys == other.keys)


class KeyStorage(ZooKeeperBase):
    log = logging.getLogger("zuul.KeyStorage")
    # /keystorage/connection/orgname
    PREFIX_PATH = "/keystorage/{}/{}"
    # /keystorage/connection/orgname/projectuniqname
    PROJECT_PATH = PREFIX_PATH + "/{}"
    SECRETS_PATH = PROJECT_PATH + "/secrets"
    SSH_PATH = PROJECT_PATH + "/ssh"
    # TODO: migrate existing keystorage to keystorage2
    TENANT_SSH_PATH = "/keystorage2/tenant/{}/ssh"

    def __init__(self, zookeeper_client, password, backup=None,
                 start_cache=True):
        super().__init__(zookeeper_client)
        self.password = password
        self.password_bytes = password.encode("utf-8")
        if start_cache:
            self.oidc_signing_keys_cache = OIDCSigningKeysCache(self.client)
        else:
            self.oidc_signing_keys_cache = None

        self.getProjectSSHKeys = cachetools.func.lru_cache(maxsize=None)(
            self.getProjectSSHKeys)
        self.getProjectSecretsKeys = cachetools.func.lru_cache(maxsize=None)(
            self.getProjectSecretsKeys)

    def createZKContext(self):
        return ZKContext(self.client, None, None, self.log)

    def stop(self):
        if self.oidc_signing_keys_cache:
            self.oidc_signing_keys_cache.stop()
        super().stop()

    def _walk(self, root):
        ret = []
        children = self.kazoo_client.get_children(root)
        if children:
            for child in children:
                path = '/'.join([root, child])
                ret.extend(self._walk(path))
        else:
            data, _ = self.kazoo_client.get(root)
            try:
                ret.append((root, json.loads(data)))
            except Exception:
                self.log.error(f"Unable to load keys at {root}")
                # Keep processing exports
        return ret

    def exportKeys(self):
        keys = {}
        for (path, data) in itertools.chain(
                self._walk('/keystorage'),
                self._walk('/keystorage2'),
        ):
            self.log.info(f"Exported: {path}")
            keys[path] = data
        return {'keys': keys}

    def importKeys(self, import_data, overwrite):
        for path, data in import_data['keys'].items():
            # This includes /keystorage2
            if not path.startswith('/keystorage'):
                self.log.error(f"Invalid path: {path}")
                return
            data = json.dumps(data, sort_keys=True).encode('utf8')
            try:
                self.kazoo_client.create(path, value=data, makepath=True)
                self.log.info(f"Created key at {path}")
            except kazoo.exceptions.NodeExistsError:
                if overwrite:
                    self.kazoo_client.set(path, value=data)
                    self.log.info(f"Updated key at {path}")
                else:
                    self.log.warning(f"Not overwriting existing key at {path}")

    def getProjectSSHKeysPath(self, connection_name, project_name):
        prefix, name = strings.unique_project_name(project_name)
        key_path = self.SSH_PATH.format(connection_name, prefix, name)
        return key_path

    def getProjectSSHKeys(self, connection_name, project_name):
        """Return the public and private keys"""
        key = self._getProjectSSHKey(connection_name, project_name)
        if key is None:
            self.log.info("Generating a new SSH key for %s/%s",
                          connection_name, project_name)
            key = paramiko.RSAKey.generate(bits=RSA_KEY_SIZE)
            key_version = 0
            key_created = int(time.time())

            try:
                self._storeProjectSSHKey(connection_name, project_name, key,
                                         key_version, key_created)
            except kazoo.exceptions.NodeExistsError:
                # Handle race condition between multiple schedulers
                # creating the same SSH key.
                key = self._getProjectSSHKey(connection_name, project_name)

        with io.StringIO() as o:
            key.write_private_key(o)
            private_key = o.getvalue()
        public_key = "ssh-rsa {}".format(key.get_base64())

        return private_key, public_key

    def loadProjectSSHKeys(self, connection_name, project_name):
        """Return the complete internal data structure"""
        key_path = self.getProjectSSHKeysPath(connection_name, project_name)
        try:
            data, _ = self.kazoo_client.get(key_path)
            return json.loads(data)
        except kazoo.exceptions.NoNodeError:
            return None

    def saveProjectSSHKeys(self, connection_name, project_name, keydata):
        """Store the complete internal data structure"""
        key_path = self.getProjectSSHKeysPath(connection_name, project_name)
        data = json.dumps(keydata, sort_keys=True).encode("utf-8")
        self.kazoo_client.create(key_path, value=data, makepath=True)

    def deleteProjectSSHKeys(self, connection_name, project_name):
        """Delete the complete internal data structure"""
        key_path = self.getProjectSSHKeysPath(connection_name, project_name)
        with suppress(kazoo.exceptions.NoNodeError):
            self.kazoo_client.delete(key_path)

    def _getProjectSSHKey(self, connection_name, project_name):
        """Load and return the public and private keys"""
        keydata = self.loadProjectSSHKeys(connection_name, project_name)
        if keydata is None:
            return None
        encrypted_key = keydata['keys'][0]["private_key"]
        with io.StringIO(encrypted_key) as o:
            return paramiko.RSAKey.from_private_key(o, self.password)

    def _storeProjectSSHKey(self, connection_name, project_name, key,
                            version, created):
        """Create the internal data structure from the key and store it"""
        # key is an rsa key object
        with io.StringIO() as o:
            key.write_private_key(o, self.password)
            private_key = o.getvalue()
        keys = [{
            "version": version,
            "created": created,
            "private_key": private_key,
        }]
        keydata = {
            'schema': 1,
            'keys': keys
        }
        self.saveProjectSSHKeys(connection_name, project_name, keydata)

    def getTenantSSHKeysPath(self, tenant_name):
        name = strings.safe_tenant_name(tenant_name)
        key_path = self.TENANT_SSH_PATH.format(name)
        return key_path

    def getTenantSSHKeys(self, tenant_name):
        """Return the public and private keys"""
        key = self._getTenantSSHKey(tenant_name)
        if key is None:
            self.log.info("Generating a new SSH key for %s",
                          tenant_name)
            key = paramiko.RSAKey.generate(bits=RSA_KEY_SIZE)
            key_version = 0
            key_created = int(time.time())

            try:
                self._storeTenantSSHKey(tenant_name, key,
                                        key_version, key_created)
            except kazoo.exceptions.NodeExistsError:
                # Handle race condition between multiple schedulers
                # creating the same SSH key.
                key = self._getTenantSSHKey(tenant_name)

        with io.StringIO() as o:
            key.write_private_key(o)
            private_key = o.getvalue()
        public_key = "ssh-rsa {}".format(key.get_base64())

        return private_key, public_key

    def loadTenantSSHKeys(self, tenant_name):
        """Return the complete internal data structure"""
        key_path = self.getTenantSSHKeysPath(tenant_name)
        try:
            data, _ = self.kazoo_client.get(key_path)
            return json.loads(data)
        except kazoo.exceptions.NoNodeError:
            return None

    def saveTenantSSHKeys(self, tenant_name, keydata):
        """Store the complete internal data structure"""
        key_path = self.getTenantSSHKeysPath(tenant_name)
        data = json.dumps(keydata, sort_keys=True).encode("utf-8")
        self.kazoo_client.create(key_path, value=data, makepath=True)

    def deleteTenantSSHKeys(self, tenant_name):
        """Delete the complete internal data structure"""
        key_path = self.getTenantSSHKeysPath(tenant_name)
        with suppress(kazoo.exceptions.NoNodeError):
            self.kazoo_client.delete(key_path)

    def _getTenantSSHKey(self, tenant_name):
        """Load and return the public and private keys"""
        keydata = self.loadTenantSSHKeys(tenant_name)
        if keydata is None:
            return None
        encrypted_key = keydata['keys'][0]["private_key"]
        with io.StringIO(encrypted_key) as o:
            return paramiko.RSAKey.from_private_key(o, self.password)

    def _storeTenantSSHKey(self, tenant_name, key,
                           version, created):
        """Create the internal data structure from the key and store it"""
        # key is an rsa key object
        with io.StringIO() as o:
            key.write_private_key(o, self.password)
            private_key = o.getvalue()
        keys = [{
            "version": version,
            "created": created,
            "private_key": private_key,
        }]
        keydata = {
            'schema': 1,
            'keys': keys
        }
        self.saveTenantSSHKeys(tenant_name, keydata)

    def getProjectSecretsKeysPath(self, connection_name, project_name):
        prefix, name = strings.unique_project_name(project_name)
        key_path = self.SECRETS_PATH.format(connection_name, prefix, name)
        return key_path

    def getProjectSecretsKeys(self, connection_name, project_name):
        """Return the public and private keys"""
        pem_private_key = self._getSecretsKey(connection_name, project_name)
        if pem_private_key is None:
            self.log.info("Generating a new secrets key for %s/%s",
                          connection_name, project_name)
            private_key, public_key = encryption.generate_rsa_keypair()
            pem_private_key = encryption.serialize_rsa_private_key(
                private_key, self.password_bytes)
            key_version = 0
            key_created = int(time.time())

            try:
                self._storeSecretsKey(connection_name, project_name,
                                      pem_private_key, key_version,
                                      key_created)
            except kazoo.exceptions.NodeExistsError:
                # Handle race condition between multiple schedulers
                # creating the same secrets key.
                pem_private_key = self._getSecretsKey(
                    connection_name, project_name)

        private_key, public_key = encryption.deserialize_rsa_keypair(
            pem_private_key, self.password_bytes)

        return private_key, public_key

    def loadProjectsSecretsKeys(self, connection_name, project_name):
        """Return the complete internal data structure"""
        key_path = self.getProjectSecretsKeysPath(
            connection_name, project_name)
        try:
            data, _ = self.kazoo_client.get(key_path)
            return json.loads(data)
        except kazoo.exceptions.NoNodeError:
            return None

    def saveProjectsSecretsKeys(self, connection_name, project_name, keydata):
        """Store the complete internal data structure"""
        key_path = self.getProjectSecretsKeysPath(
            connection_name, project_name)
        data = json.dumps(keydata, sort_keys=True).encode("utf-8")
        self.kazoo_client.create(key_path, value=data, makepath=True)

    def deleteProjectsSecretsKeys(self, connection_name, project_name):
        """Delete the complete internal data structure"""
        key_path = self.getProjectSecretsKeysPath(
            connection_name, project_name)
        with suppress(kazoo.exceptions.NoNodeError):
            self.kazoo_client.delete(key_path)

    def _getSecretsKey(self, connection_name, project_name):
        """Load and return the private key"""
        keydata = self.loadProjectsSecretsKeys(
            connection_name, project_name)
        if keydata is None:
            return None
        return keydata['keys'][0]["private_key"].encode("utf-8")

    def _storeSecretsKey(self, connection_name, project_name, key,
                         version, created):
        """Create the internal data structure from the key and store it"""
        # key is a pem-encoded (base64) private key stored in bytes
        keys = [{
            "version": version,
            "created": created,
            "private_key": key.decode("utf-8"),
        }]
        keydata = {
            'schema': 1,
            'keys': keys
        }
        self.saveProjectsSecretsKeys(connection_name, project_name, keydata)

    def deleteProjectDir(self, connection_name, project_name):
        prefix, name = strings.unique_project_name(project_name)
        project_path = self.PROJECT_PATH.format(connection_name, prefix, name)
        prefix_path = self.PREFIX_PATH.format(connection_name, prefix)
        try:
            self.kazoo_client.delete(project_path)
        except kazoo.exceptions.NotEmptyError:
            # Rely on delete only deleting empty paths by default
            self.log.warning(f"Not deleting non empty path {project_path}")
        except kazoo.exceptions.NoNodeError:
            # Already deleted
            pass
        try:
            self.kazoo_client.delete(prefix_path)
        except kazoo.exceptions.NotEmptyError:
            # Normal for the org to remain due to other projects existing.
            pass
        except kazoo.exceptions.NoNodeError:
            # Already deleted
            pass

    def rotateOidcSigningKeys(self, algorithm, rotation_interval, max_ttl):
        """
        Rotate the OIDC signing keys.
        This method is to be called periodically to rotate the OIDC
        signing keys. It creates a new key and/or remove the older
        keys when necessary.
        """
        with self.createZKContext() as context:
            OIDCSigningKeys.rotateKeys(
                context, algorithm, self.password_bytes,
                rotation_interval, max_ttl)

    def getOidcSigningKeyData(self, algorithm):
        """
        Return the key data of an algorithm of OIDC singing keys
        The data returned is from ZuulTreeCache, could be not in sync with
        the actual data in Zookeeper.
        """

        oidc_signing_keys = self.oidc_signing_keys_cache.getSigningKeys(
            algorithm)
        # If it is not found in cache, it could be the cache hasn't
        # been synced or the key has not been created yet. We need to
        # check both.
        if not oidc_signing_keys:
            with self.createZKContext() as context:
                oidc_signing_keys = OIDCSigningKeys.loadKeys(
                    context, algorithm)
                if not oidc_signing_keys:
                    OIDCSigningKeys.createAndStoreKeys(
                        context, algorithm, self.password_bytes)
                    oidc_signing_keys = OIDCSigningKeys.loadKeys(
                        context, algorithm)

        return oidc_signing_keys

    def getLatestOidcSigningKeys(self, algorithm):
        """
        Return the latest key pair of an algorithm of OIDC singing keys
        The data rerunted is from ZuulTreeCache, could be not in sync with
        the actual data in Zookeeper.
        """
        signing_key_data = self.getOidcSigningKeyData(algorithm=algorithm)
        latest_key = signing_key_data.keys[-1]
        pem_private_key = latest_key["private_key"].encode("utf-8")
        version = latest_key["version"]

        private_key, public_key = encryption.deserialize_rsa_keypair(
            pem_private_key, self.password_bytes)

        return private_key, public_key, version

    def deleteOidcSigningKeys(self, algorithm):
        """Delete the complete internal data structure"""
        with self.createZKContext() as context:
            OIDCSigningKeys.deleteKeysByAlgorithm(context, algorithm)



# Copyright (C) 2023 Acme Gating, LLC
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from ansible.plugins.action import ActionBase
from ansible.template import recursive_check_defined


class ActionModule(ActionBase):
    def run(self, tmp=None, task_vars=None):
        """This module accepts one parameter:

        _zuul_freeze_vars is a list of variable names to template.

        It stores the templated variables in a cacheable fact named
        _zuul_frozen.

        If any variable is undefined or recursively references any
        undefined variables, it is omitted from the result.

        """
        if task_vars is None:
            task_vars = dict()
        results = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        varlist = self._task.args.get('_zuul_freeze_vars')
        ret = {}
        for var in varlist:
            try:
                # Template the variable (convert_bare means treat a
                # bare variable name as {{ var }}.
                value = self._templar.template(var, convert_bare=True)
                recursive_check_defined(value)
                ret[var] = value
            except Exception:
                pass

        results['ansible_facts'] = {'_zuul_frozen': ret}
        results['_ansible_facts_cacheable'] = True

        return results



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""build_times

Revision ID: ac1dad8c9434
Revises: 3c1488fb137e
Create Date: 2023-10-31 15:27:40.014371

"""

# revision identifiers, used by Alembic.
revision = 'ac1dad8c9434'
down_revision = '3c1488fb137e'
branch_labels = None
depends_on = None

from alembic import op


BUILD_TABLE = 'zuul_build'
BUILDSET_TABLE = 'zuul_buildset'


def upgrade(table_prefix=''):
    prefixed_build = table_prefix + BUILD_TABLE
    prefixed_buildset = table_prefix + BUILDSET_TABLE
    op.create_index(
        f'{prefixed_build}_end_time_idx',
        prefixed_build, ['end_time'])
    op.create_index(
        f'{prefixed_buildset}_tenant_idx',
        prefixed_buildset, ['tenant'])


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from concurrent.futures import ThreadPoolExecutor

from kazoo.handlers.threading import SequentialThreadingHandler


class PoolSequentialThreadingHandler(SequentialThreadingHandler):
    def __init__(self):
        super().__init__()
        self._pool_executor = None

    def start(self):
        self._pool_executor = ThreadPoolExecutor(max_workers=10)
        super().start()

    def stop(self):
        super().stop()
        if self._pool_executor:
            self._pool_executor.shutdown()
        self._pool_executor = None

    def short_spawn(self, func, *args, **kwargs):
        self._pool_executor.submit(func, *args, **kwargs)



# Copyright 2017 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import configparser
import io
import logging
import json
import os
import os.path
import re
import socket
import ssl
import tempfile
import testtools
import threading
import time

import zuul.web
import zuul.lib.log_streamer
from zuul.lib.fingergw import FingerGateway
from zuul.lib.statsd import normalize_statsd_name
import tests.base
from tests.base import iterate_timeout, ZuulWebFixture, FIXTURE_DIR

import jwt
from ws4py.client import WebSocketBaseClient


class WSClient(WebSocketBaseClient):
    def __init__(self, port, build_uuid, tenant='tenant-one', token=None):
        self.port = port
        self.build_uuid = build_uuid
        self.token = token
        self.results = ''
        self.close_results = None
        self.event = threading.Event()
        uri = f'ws://[::1]:{port}/api/tenant/{tenant}/console-stream'
        super(WSClient, self).__init__(uri)

        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def received_message(self, message):
        if message.is_text:
            self.results += message.data.decode('utf-8')

    def closed(self, code, reason=None):
        self.close_results = (code, reason)
        super().closed(code, reason)

    def run(self):
        self.connect()
        req = {'uuid': self.build_uuid, 'logfile': None}
        if self.token:
            req['token'] = self.token
        self.send(json.dumps(req))
        self.event.set()
        super(WSClient, self).run()
        self.close()


class TestLogStreamer(tests.base.BaseTestCase):

    def startStreamer(self, host, port, root=None):
        self.host = host
        if not root:
            root = tempfile.gettempdir()
        return zuul.lib.log_streamer.LogStreamer(self.host, port, root)

    def test_start_stop_ipv6(self):
        streamer = self.startStreamer('::1', 0)
        self.addCleanup(streamer.stop)

        port = streamer.server.socket.getsockname()[1]
        s = socket.create_connection((self.host, port))
        s.close()

        streamer.stop()

        with testtools.ExpectedException(ConnectionRefusedError):
            s = socket.create_connection((self.host, port))
        s.close()

    def test_start_stop_ipv4(self):
        streamer = self.startStreamer('127.0.0.1', 0)
        self.addCleanup(streamer.stop)

        port = streamer.server.socket.getsockname()[1]
        s = socket.create_connection((self.host, port))
        s.close()

        streamer.stop()

        with testtools.ExpectedException(ConnectionRefusedError):
            s = socket.create_connection((self.host, port))
        s.close()


class TestStreamingBase(tests.base.AnsibleZuulTestCase):

    tenant_config_file = 'config/streamer/main.yaml'
    log = logging.getLogger("zuul.test_streaming")
    fingergw_use_ssl = False

    def setUp(self):
        super().setUp()
        self.host = '::'
        self.streamer = None
        self.stop_streamer = False
        self.streaming_data = {}
        self.test_streaming_event = threading.Event()

    def stopStreamer(self):
        self.stop_streamer = True

    def startStreamer(self, port, build_uuid, root=None):
        if not root:
            root = tempfile.gettempdir()
        self.streamer = zuul.lib.log_streamer.LogStreamer(self.host,
                                                          port, root)
        port = self.streamer.server.socket.getsockname()[1]
        s = socket.create_connection((self.host, port))
        self.addCleanup(s.close)

        req = '%s\r\n' % build_uuid
        s.sendall(req.encode('utf-8'))
        self.test_streaming_event.set()

        self.streaming_data.setdefault(None, '')
        while not self.stop_streamer:
            data = s.recv(2048)
            if not data:
                break
            self.streaming_data[None] += data.decode('utf-8')

        s.shutdown(socket.SHUT_RDWR)
        s.close()
        self.streamer.stop()

    def _readSocket(self, sock, build_uuid, event, name):
        msg = "%s\r\n" % build_uuid
        sock.sendall(msg.encode('utf-8'))
        event.set()  # notify we are connected and req sent
        while True:
            data = sock.recv(1024)
            if not data:
                break
            self.streaming_data[name] += data.decode('utf-8')
        sock.shutdown(socket.SHUT_RDWR)

    def runFingerClient(self, build_uuid, gateway_address, event, name=None):
        # Wait until the gateway is started
        for x in iterate_timeout(30, "finger client to start"):
            try:
                # NOTE(Shrews): This causes the gateway to begin to handle
                # a request for which it never receives data, and thus
                # causes the getCommand() method to timeout (seen in the
                # test results, but is harmless).
                with socket.create_connection(gateway_address) as s:
                    break
            except ConnectionRefusedError:
                pass

        self.streaming_data[name] = ''
        with socket.create_connection(gateway_address) as s:
            if self.fingergw_use_ssl:
                context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                context.verify_mode = ssl.CERT_REQUIRED
                context.check_hostname = False
                context.load_cert_chain(
                    os.path.join(FIXTURE_DIR, 'fingergw/fingergw.pem'),
                    os.path.join(FIXTURE_DIR, 'fingergw/fingergw.key'))
                context.load_verify_locations(
                    os.path.join(FIXTURE_DIR, 'fingergw/root-ca.pem'))
                with context.wrap_socket(s) as s:
                    self._readSocket(s, build_uuid, event, name)
            else:
                self._readSocket(s, build_uuid, event, name)

    def runFingerGateway(self, zone=None):
        self.log.info('Starting fingergw with zone %s', zone)
        config = configparser.ConfigParser()
        config.read_dict(self.config)
        config.read_dict({
            'fingergw': {
                'listen_address': self.host,
                'port': '0',
                'hostname': 'localhost',
            }
        })
        if zone:
            config.set('fingergw', 'zone', zone)

        if self.fingergw_use_ssl:
            self.log.info('SSL enabled for fingergw')
            config.set('fingergw', 'tls_ca',
                       os.path.join(FIXTURE_DIR, 'fingergw/root-ca.pem'))
            config.set('fingergw', 'tls_cert',
                       os.path.join(FIXTURE_DIR, 'fingergw/fingergw.pem'))
            config.set('fingergw', 'tls_key',
                       os.path.join(FIXTURE_DIR, 'fingergw/fingergw.key'))
            config.set('fingergw', 'tls_verify_hostnames', 'False')

        gateway = FingerGateway(
            config,
            command_socket=None,
            pid_file=None
        )
        gateway.history = []
        gateway.start()
        self.addCleanup(gateway.stop)

        if zone:
            for _ in iterate_timeout(20, 'fingergw is registered'):
                found = False
                for gw in self.scheds.first.sched.component_registry.\
                    all('fingergw'):
                    if gw.zone == zone and gw.public_port != 0:
                        # We need to wait for the component to be registered
                        # and for it to be listening on a valid port.
                        found = True
                        break
                if found:
                    break

        gateway_port = gateway.server.socket.getsockname()[1]
        return gateway, (self.host, gateway_port)

    def runWSClient(self, *args, **kw):
        client = WSClient(*args, **kw)
        client.event.wait()
        return client


class TestStreaming(TestStreamingBase):

    def test_streaming(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        # We don't have any real synchronization for the ansible jobs, so
        # just wait until we get our running build.
        for x in iterate_timeout(30, "builds"):
            if len(self.builds):
                break
        build = self.builds[0]
        self.assertEqual(build.name, 'python27')

        build_dir = os.path.join(self.executor_server.jobdir_root, build.uuid)
        for x in iterate_timeout(30, "build dir"):
            if os.path.exists(build_dir):
                break

        # Need to wait to make sure that jobdir gets set
        for x in iterate_timeout(30, "jobdir"):
            if build.jobdir is not None:
                break
            build = self.builds[0]

        # Wait for the job to begin running and create the ansible log file.
        # The job waits to complete until the flag file exists, so we can
        # safely access the log here. We only open it (to force a file handle
        # to be kept open for it after the job finishes) but wait to read the
        # contents until the job is done.
        ansible_log = os.path.join(build.jobdir.log_root, 'job-output.txt')
        for x in iterate_timeout(30, "ansible log"):
            if os.path.exists(ansible_log):
                break
        logfile = open(ansible_log, 'r')
        self.addCleanup(logfile.close)

        # Create a thread to stream the log. We need this to be happening
        # before we create the flag file to tell the job to complete.
        streamer_thread = threading.Thread(
            target=self.startStreamer,
            args=(0, build.uuid, self.executor_server.jobdir_root,)
        )
        streamer_thread.start()
        self.addCleanup(self.stopStreamer)
        self.test_streaming_event.wait()

        # Allow the job to complete, which should close the streaming
        # connection (and terminate the thread) as well since the log file
        # gets closed/deleted.
        flag_file = os.path.join(build_dir, 'test_wait')
        open(flag_file, 'w').close()
        self.waitUntilSettled()
        streamer_thread.join()

        # Now that the job is finished, the log file has been closed by the
        # job and deleted. However, we still have a file handle to it, so we
        # can make sure that we read the entire contents at this point.
        # Compact the returned lines into a single string for easy comparison.
        file_contents = logfile.read()
        logfile.close()

        self.log.debug("\n\nFile contents: %s\n\n", file_contents)
        self.log.debug("\n\nStreamed: %s\n\n", self.streaming_data[None])
        self.assertEqual(file_contents, self.streaming_data[None])

        # Check that we logged a multiline debug message
        pattern = (r'^\d\d\d\d-\d\d-\d\d \d\d:\d\d\:\d\d\.\d\d\d\d\d\d \| '
                   r'Debug Test Token String$')
        r = re.compile(pattern, re.MULTILINE)
        match = r.search(self.streaming_data[None])
        self.assertNotEqual(match, None)

        # Check that we logged loop_var contents properly
        pattern = r'ok: "(one|two|three)"'
        m = re.search(pattern, self.streaming_data[None])
        self.assertNotEqual(m, None)

    def test_decode_boundaries(self):
        '''
        Test multi-byte characters crossing read buffer boundaries.

        The finger client used by ZuulWeb reads in increments of 1024 bytes.
        If the last byte is a multi-byte character, we end up with an error
        similar to:

           'utf-8' codec can't decode byte 0xe2 in position 1023: \
           unexpected end of data

        By making the 1024th character in the log file a multi-byte character
        (here, the Euro character), we can test this.
        '''
        # Start the web server
        web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

        # Start the finger streamer daemon
        streamer = zuul.lib.log_streamer.LogStreamer(
            self.host, 0, self.executor_server.jobdir_root)
        self.addCleanup(streamer.stop)

        # Need to set the streaming port before submitting the job
        finger_port = streamer.server.socket.getsockname()[1]
        self.executor_server.log_streaming_port = finger_port

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        # We don't have any real synchronization for the ansible jobs, so
        # just wait until we get our running build.
        for x in iterate_timeout(30, "builds"):
            if len(self.builds):
                break
        build = self.builds[0]
        self.assertEqual(build.name, 'python27')

        build_dir = os.path.join(self.executor_server.jobdir_root, build.uuid)
        for x in iterate_timeout(30, "build dir"):
            if os.path.exists(build_dir):
                break

        # Need to wait to make sure that jobdir gets set
        for x in iterate_timeout(30, "jobdir"):
            if build.jobdir is not None:
                break
            build = self.builds[0]

        # Wait for the job to begin running and create the ansible log file.
        # The job waits to complete until the flag file exists, so we can
        # safely access the log here. We only open it (to force a file handle
        # to be kept open for it after the job finishes) but wait to read the
        # contents until the job is done.
        ansible_log = os.path.join(build.jobdir.log_root, 'job-output.txt')
        for x in iterate_timeout(30, "ansible log"):
            if os.path.exists(ansible_log):
                break

        # Replace log file contents with the 1024th character being a
        # multi-byte character.
        with io.open(ansible_log, 'w', encoding='utf8') as f:
            f.write("a" * 1023)
            f.write(u"\u20AC")

        logfile = open(ansible_log, 'r')
        self.addCleanup(logfile.close)

        # Start a thread with the websocket client
        client1 = self.runWSClient(web.port, build.uuid)
        client1.event.wait()

        # Allow the job to complete
        flag_file = os.path.join(build_dir, 'test_wait')
        open(flag_file, 'w').close()

        # Wait for the websocket client to complete, which it should when
        # it's received the full log.
        client1.thread.join()

        self.waitUntilSettled()

        file_contents = logfile.read()
        logfile.close()
        self.log.debug("\n\nFile contents: %s\n\n", file_contents)
        self.log.debug("\n\nStreamed: %s\n\n", client1.results)
        self.assertEqual(file_contents, client1.results)

        hostname = normalize_statsd_name(socket.getfqdn())
        self.assertReportedStat(
            f'zuul.web.server.{hostname}.streamers', kind='g')

    def test_websocket_streaming(self):
        # Start the web server
        web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

        # Start the finger streamer daemon
        streamer = zuul.lib.log_streamer.LogStreamer(
            self.host, 0, self.executor_server.jobdir_root)
        self.addCleanup(streamer.stop)

        # Need to set the streaming port before submitting the job
        finger_port = streamer.server.socket.getsockname()[1]
        self.executor_server.log_streaming_port = finger_port

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        # We don't have any real synchronization for the ansible jobs, so
        # just wait until we get our running build.
        for x in iterate_timeout(30, "build"):
            if len(self.builds):
                break
        build = self.builds[0]
        self.assertEqual(build.name, 'python27')

        build_dir = os.path.join(self.executor_server.jobdir_root, build.uuid)
        for x in iterate_timeout(30, "build dir"):
            if os.path.exists(build_dir):
                break

        # Need to wait to make sure that jobdir gets set
        for x in iterate_timeout(30, "jobdir"):
            if build.jobdir is not None:
                break
            build = self.builds[0]

        # Wait for the job to begin running and create the ansible log file.
        # The job waits to complete until the flag file exists, so we can
        # safely access the log here. We only open it (to force a file handle
        # to be kept open for it after the job finishes) but wait to read the
        # contents until the job is done.
        ansible_log = os.path.join(build.jobdir.log_root, 'job-output.txt')
        for x in iterate_timeout(30, "ansible log"):
            if os.path.exists(ansible_log):
                break
        logfile = open(ansible_log, 'r')
        self.addCleanup(logfile.close)

        # Start a thread with the websocket client
        client1 = self.runWSClient(web.port, build.uuid)
        client1.event.wait()
        client2 = self.runWSClient(web.port, build.uuid)
        client2.event.wait()

        # Allow the job to complete
        flag_file = os.path.join(build_dir, 'test_wait')
        open(flag_file, 'w').close()

        # Wait for the websocket client to complete, which it should when
        # it's received the full log.
        client1.thread.join()
        client2.thread.join()

        self.waitUntilSettled()

        file_contents = logfile.read()
        self.log.debug("\n\nFile contents: %s\n\n", file_contents)
        self.log.debug("\n\nStreamed: %s\n\n", client1.results)
        self.assertEqual(file_contents, client1.results)
        self.log.debug("\n\nStreamed: %s\n\n", client2.results)
        self.assertEqual(file_contents, client2.results)

    def test_websocket_hangup(self):
        # Start the web server
        web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

        # Start the finger streamer daemon
        streamer = zuul.lib.log_streamer.LogStreamer(
            self.host, 0, self.executor_server.jobdir_root)
        self.addCleanup(streamer.stop)

        # Need to set the streaming port before submitting the job
        finger_port = streamer.server.socket.getsockname()[1]
        self.executor_server.log_streaming_port = finger_port

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        # We don't have any real synchronization for the ansible jobs, so
        # just wait until we get our running build.
        for x in iterate_timeout(30, "build"):
            if len(self.builds):
                break
        build = self.builds[0]
        self.assertEqual(build.name, 'python27')

        build_dir = os.path.join(self.executor_server.jobdir_root, build.uuid)
        for x in iterate_timeout(30, "build dir"):
            if os.path.exists(build_dir):
                break

        # Need to wait to make sure that jobdir gets set
        for x in iterate_timeout(30, "jobdir"):
            if build.jobdir is not None:
                break
            build = self.builds[0]

        # Wait for the job to begin running and create the ansible log file.
        # The job waits to complete until the flag file exists, so we can
        # safely access the log here.
        ansible_log = os.path.join(build.jobdir.log_root, 'job-output.txt')
        for x in iterate_timeout(30, "ansible log"):
            if os.path.exists(ansible_log):
                break

        # Start a thread with the websocket client
        client1 = self.runWSClient(web.port, build.uuid)
        client1.event.wait()

        # Wait until we've streamed everything so far
        for x in iterate_timeout(30, "streamer is caught up"):
            with open(ansible_log, 'r') as logfile:
                if client1.results == logfile.read():
                    break
            # This is intensive, give it some time
            time.sleep(1)
        self.assertNotEqual(len(web.web.stream_manager.streamers.keys()), 0)

        # Hangup the client side
        client1.close(1000, 'test close')
        client1.thread.join()

        # The client should be de-registered shortly
        for x in iterate_timeout(30, "client cleanup"):
            if len(web.web.stream_manager.streamers.keys()) == 0:
                break

        # Allow the job to complete
        flag_file = os.path.join(build_dir, 'test_wait')
        open(flag_file, 'w').close()

        self.waitUntilSettled()

    def test_finger_gateway(self):
        # Start the finger streamer daemon
        streamer = zuul.lib.log_streamer.LogStreamer(
            self.host, 0, self.executor_server.jobdir_root)
        self.addCleanup(streamer.stop)
        finger_port = streamer.server.socket.getsockname()[1]

        # Need to set the streaming port before submitting the job
        self.executor_server.log_streaming_port = finger_port

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        # We don't have any real synchronization for the ansible jobs, so
        # just wait until we get our running build.
        for x in iterate_timeout(30, "build"):
            if len(self.builds):
                break
        build = self.builds[0]
        self.assertEqual(build.name, 'python27')

        build_dir = os.path.join(self.executor_server.jobdir_root, build.uuid)
        for x in iterate_timeout(30, "build dir"):
            if os.path.exists(build_dir):
                break

        # Need to wait to make sure that jobdir gets set
        for x in iterate_timeout(30, "jobdir"):
            if build.jobdir is not None:
                break

        # Wait for the job to begin running and create the ansible log file.
        # The job waits to complete until the flag file exists, so we can
        # safely access the log here. We only open it (to force a file handle
        # to be kept open for it after the job finishes) but wait to read the
        # contents until the job is done.
        ansible_log = os.path.join(build.jobdir.log_root, 'job-output.txt')
        for x in iterate_timeout(30, "ansible log"):
            if os.path.exists(ansible_log):
                break
        logfile = open(ansible_log, 'r')
        self.addCleanup(logfile.close)

        # Start the finger gateway daemon
        _, gateway_address = self.runFingerGateway()

        # Start a thread with the finger client
        finger_client_event = threading.Event()
        self.finger_client_results = ''
        finger_client_thread = threading.Thread(
            target=self.runFingerClient,
            args=(build.uuid, gateway_address, finger_client_event)
        )
        finger_client_thread.start()
        finger_client_event.wait()

        # Allow the job to complete
        flag_file = os.path.join(build_dir, 'test_wait')
        open(flag_file, 'w').close()

        # Wait for the finger client to complete, which it should when
        # it's received the full log.
        finger_client_thread.join()

        self.waitUntilSettled()

        file_contents = logfile.read()
        logfile.close()
        self.log.debug("\n\nFile contents: %s\n\n", file_contents)
        self.log.debug("\n\nStreamed: %s\n\n", self.streaming_data[None])
        self.assertEqual(file_contents, self.streaming_data[None])


class TestAuthWebsocketStreaming(TestStreamingBase):
    config_file = 'zuul-admin-web.conf'
    tenant_config_file = 'config/auth-streamer/main.yaml'

    def test_auth_websocket_streaming(self):
        # Start the web server
        web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

        # Start the finger streamer daemon
        streamer = zuul.lib.log_streamer.LogStreamer(
            self.host, 0, self.executor_server.jobdir_root)
        self.addCleanup(streamer.stop)

        # Need to set the streaming port before submitting the job
        finger_port = streamer.server.socket.getsockname()[1]
        self.executor_server.log_streaming_port = finger_port

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        # We don't have any real synchronization for the ansible jobs, so
        # just wait until we get our running build.
        for x in iterate_timeout(30, "build"):
            if len(self.builds):
                break
        build = self.builds[0]
        self.assertEqual(build.name, 'python27')

        build_dir = os.path.join(self.executor_server.jobdir_root, build.uuid)
        for x in iterate_timeout(30, "build dir"):
            if os.path.exists(build_dir):
                break

        # Need to wait to make sure that jobdir gets set
        for x in iterate_timeout(30, "jobdir"):
            if build.jobdir is not None:
                break
            build = self.builds[0]

        # Wait for the job to begin running and create the ansible log file.
        # The job waits to complete until the flag file exists, so we can
        # safely access the log here. We only open it (to force a file handle
        # to be kept open for it after the job finishes) but wait to read the
        # contents until the job is done.
        ansible_log = os.path.join(build.jobdir.log_root, 'job-output.txt')
        for x in iterate_timeout(30, "ansible log"):
            if os.path.exists(ansible_log):
                break
        logfile = open(ansible_log, 'r')
        self.addCleanup(logfile.close)

        # Attempt to stream the log without an authz token
        client1 = self.runWSClient(web.port, build.uuid)
        client1.thread.join()
        self.assertEqual(
            (4000, b'MissingAuthError'),
            client1.close_results)

        # Attempt to bypass authz by using a valid token with a
        # different tenant
        authz = {'iss': 'zuul_operator',
                 'aud': 'zuul.example.com',
                 'sub': 'testuser',
                 'groups': ['users'],
                 'exp': int(time.time()) + 3600}
        token = jwt.encode(authz, key='ThisIsABadSecretOnlyUsedForTesting',
                           algorithm='HS256')
        client2 = self.runWSClient(web.port, build.uuid, tenant='tenant-two',
                                   token=token)
        client2.thread.join()
        self.assertEqual(
            (4011, b'Build not found'),
            client2.close_results)

        # Finally, use a valid token
        client3 = self.runWSClient(web.port, build.uuid, token=token)

        # Allow the job to complete
        flag_file = os.path.join(build_dir, 'test_wait')
        open(flag_file, 'w').close()

        # Wait for the websocket client to complete, which it should when
        # it's received the full log.
        client3.thread.join()

        self.waitUntilSettled()

        file_contents = logfile.read()
        self.log.debug("\n\nFile contents: %s\n\n", file_contents)
        self.log.debug("\n\nStreamed: %s\n\n", client3.results)
        self.assertEqual(file_contents, client3.results)


class CountingFingerRequestHandler(zuul.lib.fingergw.RequestHandler):

    def _fingerClient(self, server, port, build_uuid, use_ssl):
        self.fingergw.history.append(build_uuid)
        super()._fingerClient(server, port, build_uuid, use_ssl)


class TestStreamingZones(TestStreamingBase):

    def setUp(self):
        super().setUp()
        self.fake_nodepool.attributes = {'executor-zone': 'eu-central'}
        zuul.lib.fingergw.FingerGateway.handler_class = \
            CountingFingerRequestHandler

    def setup_config(self, config_file: str):
        config = super().setup_config(config_file)
        config.set('executor', 'zone', 'eu-central')
        return config

    def _run_finger_client(self, build, address, name):
        # Start a thread with the finger client
        finger_client_event = threading.Event()
        self.finger_client_results = ''
        finger_client_thread = threading.Thread(
            target=self.runFingerClient,
            args=(build.uuid, address, finger_client_event),
            kwargs={'name': name}
        )
        finger_client_thread.start()
        finger_client_event.wait()
        return finger_client_thread

    def test_finger_gateway(self):
        # Start the finger streamer daemon
        streamer = zuul.lib.log_streamer.LogStreamer(
            self.host, 0, self.executor_server.jobdir_root)
        self.addCleanup(streamer.stop)
        finger_port = streamer.server.socket.getsockname()[1]

        # Need to set the streaming port before submitting the job
        self.executor_server.log_streaming_port = finger_port

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        # We don't have any real synchronization for the ansible jobs, so
        # just wait until we get our running build.
        for x in iterate_timeout(30, "build"):
            if len(self.builds):
                break
        build = self.builds[0]
        self.assertEqual(build.name, 'python27')

        build_dir = os.path.join(self.executor_server.jobdir_root, build.uuid)
        for x in iterate_timeout(30, "build dir"):
            if os.path.exists(build_dir):
                break

        # Need to wait to make sure that jobdir gets set
        for x in iterate_timeout(30, "jobdir"):
            if build.jobdir is not None:
                break

        # Wait for the job to begin running and create the ansible log file.
        # The job waits to complete until the flag file exists, so we can
        # safely access the log here. We only open it (to force a file handle
        # to be kept open for it after the job finishes) but wait to read the
        # contents until the job is done.
        ansible_log = os.path.join(build.jobdir.log_root, 'job-output.txt')
        for x in iterate_timeout(30, "ansible log"):
            if os.path.exists(ansible_log):
                break
        logfile = open(ansible_log, 'r')
        self.addCleanup(logfile.close)

        def wait_for_stream(name):
            for x in iterate_timeout(30, "incoming streaming data"):
                if len(self.streaming_data.get(name, '')) > 0:
                    break

        # Start the finger gateway daemons
        try:
            gateway_unzoned, gateway_unzoned_address = self.runFingerGateway()
            gateway_us_west, gateway_us_west_address = self.runFingerGateway(
                zone='us-west')
        except Exception:
            self.log.exception("Failed to run finger gateway")
            raise

        # This finger client runs against a finger gateway in a different zone
        # while there is no gateway in the worker zone yet. This should work.
        finger_client_us_west_alone = self._run_finger_client(
            build, gateway_us_west_address, name='us-west-alone')
        # The stream must go only via gateway_us_west
        wait_for_stream('us-west-alone')
        self.assertEqual(0, len(gateway_unzoned.history))
        self.assertEqual(1, len(gateway_us_west.history))
        gateway_unzoned.history.clear()
        gateway_us_west.history.clear()

        # This finger client runs against an unzoned finger gateway
        finger_client_unzoned = self._run_finger_client(
            build, gateway_unzoned_address, name='unzoned')
        wait_for_stream('unzoned')
        self.assertEqual(1, len(gateway_unzoned.history))
        self.assertEqual(0, len(gateway_us_west.history))
        gateway_unzoned.history.clear()
        gateway_us_west.history.clear()

        # Now start a finger gateway in the target zone.
        gateway_eu_central, gateway_eu_central_address = self.runFingerGateway(
            zone='eu-central')

        # This finger client runs against a finger gateway in a different zone
        # while there is a gateway in the worker zone. This should route via
        # the gateway in the worker zone.
        finger_client_us_west = self._run_finger_client(
            build, gateway_us_west_address, name='us-west')
        # The stream must go only via gateway_us_west
        wait_for_stream('us-west')
        self.assertEqual(0, len(gateway_unzoned.history))
        self.assertEqual(1, len(gateway_eu_central.history))
        self.assertEqual(1, len(gateway_us_west.history))
        gateway_unzoned.history.clear()
        gateway_eu_central.history.clear()
        gateway_us_west.history.clear()

        # This finger client runs against an unzoned finger gateway
        # while there is a gateway in the worker zone. It should still
        # route via the gateway in the worker zone since that may be
        # the only way it's accessible.
        finger_client_unzoned2 = self._run_finger_client(
            build, gateway_unzoned_address, name='unzoned2')
        wait_for_stream('unzoned2')
        self.assertEqual(1, len(gateway_unzoned.history))
        self.assertEqual(1, len(gateway_eu_central.history))
        self.assertEqual(0, len(gateway_us_west.history))
        gateway_unzoned.history.clear()
        gateway_eu_central.history.clear()
        gateway_us_west.history.clear()

        # This finger client runs against the target finger gateway.
        finger_client_eu_central = self._run_finger_client(
            build, gateway_eu_central_address, name='eu-central')
        wait_for_stream('eu-central')
        self.assertEqual(0, len(gateway_unzoned.history))
        self.assertEqual(1, len(gateway_eu_central.history))
        self.assertEqual(0, len(gateway_us_west.history))
        gateway_unzoned.history.clear()
        gateway_eu_central.history.clear()
        gateway_us_west.history.clear()

        # Allow the job to complete
        flag_file = os.path.join(build_dir, 'test_wait')
        open(flag_file, 'w').close()

        # Wait for the finger client to complete, which it should when
        # it's received the full log.
        finger_client_us_west_alone.join()
        finger_client_us_west.join()
        finger_client_eu_central.join()
        finger_client_unzoned.join()
        finger_client_unzoned2.join()

        self.waitUntilSettled()

        file_contents = logfile.read()
        logfile.close()
        self.log.debug("\n\nFile contents: %s\n\n", file_contents)
        self.log.debug("\n\nStreamed: %s\n\n",
                       self.streaming_data['us-west-alone'])
        self.assertEqual(file_contents, self.streaming_data['us-west-alone'])
        self.assertEqual(file_contents, self.streaming_data['us-west'])
        self.assertEqual(file_contents, self.streaming_data['unzoned'])
        self.assertEqual(file_contents, self.streaming_data['unzoned2'])
        self.assertEqual(file_contents, self.streaming_data['eu-central'])


class TestStreamingZonesSSL(TestStreamingZones):
    fingergw_use_ssl = True


class TestStreamingUnzonedJob(TestStreamingZones):

    def setUp(self):
        super().setUp()
        self.fake_nodepool.attributes = None

    def setup_config(self, config_file: str):
        config = super().setup_config(config_file)
        config.set('executor', 'allow_unzoned', 'true')
        return config



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul import model
from zuul.manager.shared import SharedQueuePipelineManager


class SerialPipelineManager(SharedQueuePipelineManager):
    """PipelineManager with shared queues and a window of 1"""

    changes_merge = False
    type = 'serial'

    def constructChangeQueue(self, queue_name):
        return model.ChangeQueue.new(
            self.current_context,
            manager=self,
            window=1,
            window_floor=1,
            window_ceiling=1,
            window_increase_type='none',
            window_decrease_type='none',
            name=queue_name)

    def dequeueItem(self, item):
        super(SerialPipelineManager, self).dequeueItem(item)
        # If this was a dynamic queue from a speculative change,
        # remove the queue (if empty)
        if item.queue.dynamic:
            if not item.queue.queue:
                self.state.removeQueue(item.queue)



#!/usr/bin/env python3
# Copyright 2013 OpenStack Foundation
# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen
import json
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('url', help='The URL of the running Zuul instance')
parser.add_argument('tenant', help='The Zuul tenant', nargs='?')
parser.add_argument('pipeline', help='The name of the Zuul pipeline',
                    nargs='?')
parser.add_argument('--use-config',
                    metavar='CONFIG',
                    help='The name of the zuul-client config to use')
options = parser.parse_args()

command = 'zuul-client'
if options.use_config:
    command += f' --use-config {options.use_config}'

# Check if tenant is white label
info = json.loads(urlopen('%s/api/info' % options.url).read())
api_tenant = info.get('info', {}).get('tenant')
tenants = []
if api_tenant:
    if api_tenant == options.tenant:
        tenants.append(None)
    else:
        print("Error: %s doesn't match tenant %s (!= %s)" % (
            options.url, options.tenant, api_tenant))
        exit(1)
else:
    tenants_url = '%s/api/tenants' % options.url
    data = json.loads(urlopen(tenants_url).read())
    for tenant in data:
        tenants.append(tenant['name'])

for tenant in tenants:
    if tenant is None:
        status_url = '%s/api/status' % options.url
    else:
        status_url = '%s/api/tenant/%s/status' % (options.url, tenant)

    data = json.loads(urlopen(status_url).read())

    for pipeline in data['pipelines']:
        if options.pipeline and pipeline['name'] != options.pipeline:
            continue
        for queue in pipeline.get('change_queues', []):
            for head in queue['heads']:
                for change in head:
                    if not change['live']:
                        continue

                    if change['id'] and ',' in change['id']:
                        # change triggered
                        cid, cps = change['id'].split(',')
                        print("%s enqueue"
                              " --tenant %s"
                              " --pipeline %s"
                              " --project %s"
                              " --change %s,%s" % (command, tenant,
                                                   pipeline['name'],
                                                   change['project_canonical'],
                                                   cid, cps))
                    else:
                        # ref triggered
                        cmd = '%s enqueue-ref' \
                              ' --tenant %s' \
                              ' --pipeline %s' \
                              ' --project %s' \
                              ' --ref %s' % (command, tenant,
                                             pipeline['name'],
                                             change['project_canonical'],
                                             change['ref'])
                        if change['id']:
                            cmd += ' --newrev %s' % change['id']
                        print(cmd)



# Copyright (c) 2017 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import ZuulTestCase, simple_layout


class TestGithubCrossRepoDeps(ZuulTestCase):
    """Test Github cross-repo dependencies"""
    config_file = 'zuul-github-driver.conf'
    scheduler_count = 1

    @simple_layout('layouts/crd-github.yaml', driver='github')
    def test_crd_independent(self):
        "Test cross-repo dependences on an independent pipeline"

        # Create a change in project1 that a project2 change will depend on
        A = self.fake_github.openFakePullRequest('org/project1', 'master', 'A')

        # Create a commit in B that sets the dependency on A
        msg = "Depends-On: https://github.com/org/project1/pull/%s" % A.number
        B = self.fake_github.openFakePullRequest('org/project2', 'master', 'B',
                                                 body=msg)

        # Make an event to re-use
        event = B.getPullRequestEditedEvent()

        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        # The changes for the job from project2 should include the project1
        # PR contet
        changes = self.getJobFromHistory(
            'project2-test', 'org/project2').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.head_sha,
                                                   B.number,
                                                   B.head_sha))

        # There should be no more changes in the queue
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

    @simple_layout('layouts/crd-github.yaml', driver='github')
    def test_crd_dependent(self):
        "Test cross-repo dependences on a dependent pipeline"

        # Create a change in project3 that a project4 change will depend on
        A = self.fake_github.openFakePullRequest('org/project3', 'master', 'A')

        # Create a commit in B that sets the dependency on A
        msg = "Depends-On: https://github.com/org/project3/pull/%s" % A.number
        B = self.fake_github.openFakePullRequest('org/project4', 'master', 'B',
                                                 body=msg)

        # Make an event to re-use
        event = B.getPullRequestEditedEvent()

        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        # The changes for the job from project4 should include the project3
        # PR contet
        changes = self.getJobFromHistory(
            'project4-test', 'org/project4').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.head_sha,
                                                   B.number,
                                                   B.head_sha))

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)

    @simple_layout('layouts/crd-github.yaml', driver='github')
    def test_crd_unshared_dependent(self):
        "Test cross-repo dependences on unshared dependent pipeline"

        # Create a change in project1 that a project2 change will depend on
        A = self.fake_github.openFakePullRequest('org/project5', 'master', 'A')

        # Create a commit in B that sets the dependency on A
        msg = "Depends-On: https://github.com/org/project5/pull/%s" % A.number
        B = self.fake_github.openFakePullRequest('org/project6', 'master', 'B',
                                                 body=msg)

        # Make an event for B
        event = B.getPullRequestEditedEvent()

        # Emit for B, which should not enqueue A because they do not share
        # A queue. Since B depends on A, and A isn't enqueue, B will not run
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        self.assertEqual(0, len(self.history))

        # Enqueue A alone, let it finish
        self.fake_github.emitEvent(A.getPullRequestEditedEvent())
        self.waitUntilSettled()

        self.assertTrue(A.is_merged)
        self.assertFalse(B.is_merged)
        self.assertEqual(1, len(self.history))

        # With A merged, B should go through
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        self.assertTrue(B.is_merged)
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/crd-github.yaml', driver='github')
    def test_crd_cycle(self):
        "Test cross-repo dependency cycles"

        # A -> B -> A
        msg = "Depends-On: https://github.com/org/project6/pull/2"
        A = self.fake_github.openFakePullRequest('org/project5', 'master', 'A',
                                                 body=msg)
        msg = "Depends-On: https://github.com/org/project5/pull/1"
        B = self.fake_github.openFakePullRequest('org/project6', 'master', 'B',
                                                 body=msg)

        self.fake_github.emitEvent(A.getPullRequestEditedEvent())
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertFalse(B.is_merged)
        self.assertEqual(0, len(self.history))

    @simple_layout('layouts/crd-github.yaml', driver='github')
    def test_crd_needed_changes(self):
        "Test cross-repo needed changes discovery"

        # Given change A and B, where B depends on A, when A
        # completes B should be enqueued (using a shared queue)

        # Create a change in project3 that a project4 change will depend on
        A = self.fake_github.openFakePullRequest('org/project3', 'master', 'A')

        # Set B to depend on A
        msg = "Depends-On: https://github.com/org/project3/pull/%s" % A.number
        B = self.fake_github.openFakePullRequest('org/project4', 'master', 'B',
                                                 body=msg)

        # Enqueue A, which when finished should enqueue B
        self.fake_github.emitEvent(A.getPullRequestEditedEvent())
        self.waitUntilSettled()

        # The changes for the job from project4 should include the project3
        # PR contet
        changes = self.getJobFromHistory(
            'project4-test', 'org/project4').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.head_sha,
                                                   B.number,
                                                   B.head_sha))

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)

    @simple_layout('layouts/github-message-update.yaml', driver='github')
    def test_crd_message_update(self):
        "Test a change is dequeued when the PR in its depends-on is updated"

        # Create a change in project1 that a project2 change will depend on
        A = self.fake_github.openFakePullRequest('org/project1', 'master', 'A')

        # Create a commit in B that sets the dependency on A
        msg = "Depends-On: https://github.com/org/project1/pull/%s" % A.number
        B = self.fake_github.openFakePullRequest('org/project2', 'master', 'B',
                                                 body=msg)

        # Create a commit in C that sets the dependency on B
        msg = "Depends-On: https://github.com/org/project2/pull/%s" % B.number
        C = self.fake_github.openFakePullRequest('org/project3', 'master', 'C',
                                                 body=msg)

        # A change we'll use later to replace A
        A1 = self.fake_github.openFakePullRequest(
            'org/project1', 'master', 'A1')

        self.executor_server.hold_jobs_in_build = True

        # Enqueue A,B,C
        self.fake_github.emitEvent(C.getReviewAddedEvent('approved'))
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)
        items = self.getAllItems('tenant-one', 'check')
        self.assertEqual(len(items), 3)

        # Update B to point at A1 instead of A
        msg = "Depends-On: https://github.com/org/project1/pull/%s" % A1.number
        self.fake_github.emitEvent(B.editBody(msg))
        self.waitUntilSettled()

        # Release
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        # The job should be aborted since B was updated while enqueued.
        self.assertHistory([dict(name='project3-test', result='ABORTED')])

    @simple_layout('layouts/crd-github.yaml', driver='github')
    def test_crd_dependent_close_reopen(self):
        """
        Test that closing and reopening PR correctly re-enqueues the
        dependent change in the correct order.
        """
        self.executor_server.hold_jobs_in_build = True

        A = self.fake_github.openFakePullRequest('org/project3', 'master', 'A')
        B = self.fake_github.openFakePullRequest(
            'org/project4', 'master', 'B',
            body=f"Depends-On: https://github.com/org/project3/pull/{A.number}"
        )

        self.fake_github.emitEvent(B.getPullRequestEditedEvent())
        self.waitUntilSettled()

        # Close and reopen PR A
        self.fake_github.emitEvent(A.getPullRequestClosedEvent())
        self.fake_github.emitEvent(A.getPullRequestReopenedEvent())
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        # The changes for the job from project4 should include the project3
        # PR content
        self.assertHistory([
            dict(name='project3-test', result='ABORTED',
                 changes=f"{A.number},{A.head_sha}"),
            dict(name='project4-test', result='ABORTED',
                 changes=f"{A.number},{A.head_sha} {B.number},{B.head_sha}"),
            dict(name='project3-test', result='SUCCESS',
                 changes=f"{A.number},{A.head_sha}"),
            dict(name='project4-test', result='SUCCESS',
                 changes=f"{A.number},{A.head_sha} {B.number},{B.head_sha}"),
        ], ordered=False)

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)



# Copyright 2015 Rackspace Australia
# Copyright 2023-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import textwrap

# Error severity
SEVERITY_ERROR = 'error'
SEVERITY_WARNING = 'warning'


class ChangeNotFound(Exception):
    def __init__(self, number, ps):
        self.number = number
        self.ps = ps
        self.change = "%s,%s" % (str(number), str(ps))
        message = "Change %s not found" % self.change
        super(ChangeNotFound, self).__init__(message)


class RevNotFound(Exception):
    def __init__(self, project, rev):
        self.project = project
        self.revision = rev
        message = ("Failed to checkout project '%s' at revision '%s'"
                   % (self.project, self.revision))
        super(RevNotFound, self).__init__(message)


class MergeFailure(Exception):
    pass


class MissingBuildsetError(Exception):
    pass


class ConfigurationError(Exception):
    pass


class StreamingError(Exception):
    pass


class DependencyLimitExceededError(Exception):
    pass


class VariableNameError(Exception):
    pass


# Provider exceptions
class LaunchStatusException(Exception):
    statsd_key = 'error.status'


class LaunchNetworkException(Exception):
    statsd_key = 'error.network'


class LaunchKeyscanException(Exception):
    statsd_key = 'error.keyscan'


class CapacityException(Exception):
    statsd_key = 'error.capacity'


class TimeoutException(Exception):
    pass


class ConnectionTimeoutException(TimeoutException):
    statsd_key = 'error.ssh'


class RuntimeConfigurationException(Exception):
    pass


class QuotaException(Exception):
    pass


class AlgorithmNotSupportedException(Exception):
    pass


# Authentication Exceptions
class AuthTokenException(Exception):
    defaultMsg = 'Unknown Error'
    HTTPError = 400

    def __init__(self, realm=None, msg=None):
        super(AuthTokenException, self).__init__(msg or self.defaultMsg)
        self.realm = realm
        self.error = self.__class__.__name__
        self.error_description = msg or self.defaultMsg

    def getAdditionalHeaders(self):
        return {}


class JWKSException(AuthTokenException):
    defaultMsg = 'Unknown error involving JSON Web Key Set'


class AuthTokenForbiddenException(AuthTokenException):
    defaultMsg = 'Insufficient privileges'
    HTTPError = 403


class AuthTokenUnauthorizedException(AuthTokenException):
    defaultMsg = 'This action requires authentication'
    HTTPError = 401

    def getAdditionalHeaders(self):
        error_header = '''Bearer realm="%s"
       error="%s"
       error_description="%s"'''
        return {"WWW-Authenticate": error_header % (self.realm,
                                                    self.error,
                                                    self.error_description)}


class AuthTokenUndecodedException(AuthTokenUnauthorizedException):
    defaultMsg = 'Auth Token could not be decoded'


class AuthTokenInvalidSignatureException(AuthTokenUnauthorizedException):
    defaultMsg = 'Invalid signature'


class BearerTokenRequiredError(AuthTokenUnauthorizedException):
    defaultMsg = 'Authorization with bearer token required'


class IssuerUnknownError(AuthTokenUnauthorizedException):
    defaultMsg = 'Issuer unknown'


class MissingClaimError(AuthTokenUnauthorizedException):
    defaultMsg = 'Token is missing claims'


class IncorrectAudienceError(AuthTokenUnauthorizedException):
    defaultMsg = 'Incorrect audience'


class TokenExpiredError(AuthTokenUnauthorizedException):
    defaultMsg = 'Token has expired'


class MissingUIDClaimError(MissingClaimError):
    defaultMsg = 'Token is missing id claim'


class IncorrectZuulAdminClaimError(AuthTokenUnauthorizedException):
    defaultMsg = (
        'The "zuul.admin" claim is expected to be a list of tenants')


class UnauthorizedZuulAdminClaimError(AuthTokenUnauthorizedException):
    defaultMsg = 'Issuer is not allowed to set "zuul.admin" claim'


class ConfigurationSyntaxError(Exception):
    zuul_error_name = 'Unknown Configuration Error'
    zuul_error_severity = SEVERITY_ERROR


class NodeFromGroupNotFoundError(ConfigurationSyntaxError):
    zuul_error_name = 'Node From Group Not Found'

    def __init__(self, nodeset, node, group):
        message = textwrap.dedent("""\
        In {nodeset} the group "{group}" contains a
        node named "{node}" which is not defined in the nodeset.""")
        message = textwrap.fill(message.format(nodeset=nodeset,
                                               node=node, group=group))
        super(NodeFromGroupNotFoundError, self).__init__(message)


class DuplicateNodeError(ConfigurationSyntaxError):
    zuul_error_name = 'Duplicate Node'

    def __init__(self, nodeset, node):
        message = textwrap.dedent("""\
        In nodeset "{nodeset}" the node "{node}" appears multiple times.
        Node names must be unique within a nodeset.""")
        message = textwrap.fill(message.format(nodeset=nodeset,
                                               node=node))
        super(DuplicateNodeError, self).__init__(message)


class UnknownConnection(ConfigurationSyntaxError):
    zuul_error_name = 'Unknown Connection'

    def __init__(self, connection_name):
        message = textwrap.dedent("""\
        Unknown connection named "{connection}".""")
        message = textwrap.fill(message.format(connection=connection_name))
        super(UnknownConnection, self).__init__(message)


class LabelForbiddenError(ConfigurationSyntaxError):
    zuul_error_name = 'Label Forbidden'

    def __init__(self, label, allowed_labels, disallowed_labels):
        message = textwrap.dedent("""\
        Label named "{label}" is not part of the allowed
        labels ({allowed_labels}) for this tenant.""")
        # Make a string that looks like "a, b and not c, d" if we have
        # both allowed and disallowed labels.
        labels = ", ".join(allowed_labels or [])
        if allowed_labels and disallowed_labels:
            labels += ' and '
        if disallowed_labels:
            labels += 'not '
            labels += ", ".join(disallowed_labels)
        message = textwrap.fill(message.format(
            label=label,
            allowed_labels=labels))
        super(LabelForbiddenError, self).__init__(message)


class MaxTimeoutError(ConfigurationSyntaxError):
    zuul_error_name = 'Max Timeout Exceeded'

    def __init__(self, job, tenant):
        message = textwrap.dedent("""\
        The job "{job}" exceeds tenant max-job-timeout {maxtimeout}.""")
        message = textwrap.fill(message.format(
            job=job.name, maxtimeout=tenant.max_job_timeout))
        super(MaxTimeoutError, self).__init__(message)


class PreTimeoutExceedsTimeoutError(ConfigurationSyntaxError):
    zuul_error_name = 'Pre-Timeout Exceeds Timeout'

    def __init__(self, job):
        message = textwrap.dedent("""\
        The job "{job}" has a pre-timeout of {pre_timeout}
        that exceeds its timeout of {timeout}.""")
        message = textwrap.fill(message.format(
            job=job.name, pre_timeout=job.pre_timeout, timeout=job.timeout))
        super(PreTimeoutExceedsTimeoutError, self).__init__(message)


class MaxOIDCTTLError(ConfigurationSyntaxError):
    zuul_error_name = 'Max OIDC TTL Exceeded'

    def __init__(self, secret, tenant):
        message = textwrap.dedent("""\
        The oidc secret "{secret}" exceeds tenant max-oidc-ttl {max_ttl}.""")
        message = textwrap.fill(message.format(
            secret=secret.name, max_ttl=tenant.max_oidc_ttl))
        super(MaxOIDCTTLError, self).__init__(message)


class OIDCIssuerNotAllowedError(ConfigurationSyntaxError):
    zuul_error_name = 'OIDC issuer not allowed'

    def __init__(self, secret, issuer):
        message = textwrap.dedent("""\
        The iss "{issuer}" in oidc secret "{secret}" is not allowed.""")
        message = textwrap.fill(message.format(
            secret=secret.name, issuer=issuer))
        super(OIDCIssuerNotAllowedError, self).__init__(message)


class DuplicateGroupError(ConfigurationSyntaxError):
    zuul_error_name = 'Duplicate Nodeset Group'

    def __init__(self, nodeset, group):
        message = textwrap.dedent("""\
        In {nodeset} the group "{group}" appears multiple times.
        Group names must be unique within a nodeset.""")
        message = textwrap.fill(message.format(nodeset=nodeset,
                                               group=group))
        super(DuplicateGroupError, self).__init__(message)


class ReporterJobNotPermittedError(ConfigurationSyntaxError):
    zuul_error_name = 'Reporter Job Not Permitted'

    def __init__(self, project):
        message = textwrap.dedent(f"""\
        The project {project} is not permitted to configure a
        reporter job.""")
        message = textwrap.fill(message)
        super().__init__(message)


class ProjectNotFoundError(ConfigurationSyntaxError):
    zuul_error_name = 'Project Not Found'

    def __init__(self, project):
        projects = None
        if isinstance(project, (list, tuple)):
            if len(project) > 1:
                projects = ', '.join(f'"{p}"' for p in project)
            else:
                project = project[0]
        if projects:
            message = textwrap.dedent(f"""\
            The projects {projects} were not found.  All projects
            referenced within a Zuul configuration must first be
            added to the main configuration file by the Zuul
            administrator.""")
        else:
            message = textwrap.dedent(f"""\
            The project "{project}" was not found.  All projects
            referenced within a Zuul configuration must first be
            added to the main configuration file by the Zuul
            administrator.""")
        message = textwrap.fill(message)
        super(ProjectNotFoundError, self).__init__(message)


class TemplateNotFoundError(ConfigurationSyntaxError):
    zuul_error_name = 'Template Not Found'

    def __init__(self, template):
        message = textwrap.dedent("""\
        The project template "{template}" was not found.
        """)
        message = textwrap.fill(message.format(template=template))
        super(TemplateNotFoundError, self).__init__(message)


class NodesetNotFoundError(ConfigurationSyntaxError):
    zuul_error_name = 'Nodeset Not Found'

    def __init__(self, nodeset):
        message = textwrap.dedent("""\
        The nodeset "{nodeset}" was not found.
        """)
        message = textwrap.fill(message.format(nodeset=nodeset))
        super(NodesetNotFoundError, self).__init__(message)


class LabelNotFoundError(ConfigurationSyntaxError):
    zuul_error_name = 'Label Not Found'

    def __init__(self, label):
        message = textwrap.dedent("""\
        The label "{label}" was not found.
        """)
        message = textwrap.fill(message.format(label=label))
        super(LabelNotFoundError, self).__init__(message)


class PipelineNotPermittedError(ConfigurationSyntaxError):
    zuul_error_name = 'Pipeline Forbidden'

    def __init__(self):
        message = textwrap.dedent("""\
        Pipelines may not be defined in untrusted repos,
        they may only be defined in config repos.""")
        message = textwrap.fill(message)
        super(PipelineNotPermittedError, self).__init__(message)


class ProjectNotPermittedError(ConfigurationSyntaxError):
    zuul_error_name = 'Project Forbidden'

    def __init__(self):
        message = textwrap.dedent("""\
        Within an untrusted project, the only project definition
        permitted is that of the project itself.""")
        message = textwrap.fill(message)
        super(ProjectNotPermittedError, self).__init__(message)


class GlobalSemaphoreNotFoundError(ConfigurationSyntaxError):
    zuul_error_name = 'Global Semaphore Not Found'

    def __init__(self, semaphore):
        message = textwrap.dedent("""\
        The global semaphore "{semaphore}" was not found.  All
        global semaphores must be added to the main configuration
        file by the Zuul administrator.""")
        message = textwrap.fill(message.format(semaphore=semaphore))
        super(GlobalSemaphoreNotFoundError, self).__init__(message)


class YAMLDuplicateKeyError(ConfigurationSyntaxError):
    def __init__(self, key, source_context, start_mark):
        self.source_context = source_context
        self.start_mark = start_mark
        message = (f'The key "{key}" appears more than once; '
                   'duplicate keys are not permitted.')
        super(YAMLDuplicateKeyError, self).__init__(message)


class ConfigurationSyntaxWarning:
    zuul_error_name = 'Unknown Configuration Warning'
    zuul_error_severity = SEVERITY_WARNING
    zuul_error_message = 'Unknown Configuration Warning'

    def __init__(self, message=None):
        if message:
            self.zuul_error_message = message


class MultipleProjectConfigurations(ConfigurationSyntaxWarning):
    zuul_error_name = 'Multiple Project Configurations'
    zuul_error_problem = 'configuration error'

    def __init__(self, source_context):
        message = textwrap.dedent(f"""\
        Configuration in {source_context.path} ignored because project-branch
        is already configured.""")
        message = textwrap.fill(message)
        super().__init__(message)


class DeprecationWarning(ConfigurationSyntaxWarning):
    zuul_error_problem = 'deprecated syntax'


class RegexDeprecation(DeprecationWarning):
    zuul_error_name = 'Regex Deprecation'
    zuul_error_message = """\
All regular expressions must conform to RE2 syntax, but an
expression using the deprecated Perl-style syntax has been detected.
Adjust the configuration to conform to RE2 syntax."""

    def __init__(self, message=None):
        if message:
            message = (self.zuul_error_message +
                       f"\n\nThe RE2 syntax error is: {message}")
        super().__init__(message)


class CleanupRunDeprecation(DeprecationWarning):
    zuul_error_name = 'Cleanup Run Deprecation'
    zuul_error_message = """\
The cleanup-run job attribute is deprecated.  Replace it with
post-run playbooks with the `cleanup` attribute set."""



# Copyright 2021, 2023-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul import model
from zuul.lib.logutil import get_annotated_logger
from zuul.manager import PipelineManager, DynamicChangeQueueContextManager


class SupercedentPipelineManager(PipelineManager):
    """PipelineManager with one queue per project and a window of 1"""

    changes_merge = False
    type = 'supercedent'

    def getChangeQueue(self, change, event, existing=None):
        log = get_annotated_logger(self.log, event)

        # creates a new change queue for every project-ref
        # combination.
        if existing:
            return DynamicChangeQueueContextManager(existing)

        # Don't use Pipeline.getQueue to find an existing queue
        # because we're matching project and (branch or ref).
        for queue in self.state.queues:
            if (queue.queue[-1].changes[0].project == change.project and
                ((hasattr(change, 'branch') and
                  hasattr(queue.queue[-1].changes[0], 'branch') and
                  queue.queue[-1].changes[0].branch == change.branch) or
                queue.queue[-1].changes[0].ref == change.ref)):
                log.debug("Found existing queue %s", queue)
                return DynamicChangeQueueContextManager(queue)
        change_queue = model.ChangeQueue.new(
            self.current_context,
            manager=self,
            window=1,
            window_floor=1,
            window_ceiling=1,
            window_increase_type='none',
            window_decrease_type='none',
            dynamic=True)
        change_queue.addProject(change.project, None)
        self.state.addQueue(change_queue)
        log.debug("Dynamically created queue %s", change_queue)
        return DynamicChangeQueueContextManager(
            change_queue, allow_delete=True)

    def _pruneQueues(self):
        # Leave the first item in the queue, as it's running, and the
        # last item, as it's the most recent, but remove any items in
        # between.  This is what causes the last item to "supercede"
        # any previously enqueued items (which we know aren't running
        # jobs because the window size is 1).
        for queue in self.state.queues[:]:
            remove = queue.queue[1:-1]
            for item in remove:
                self.log.debug("Item %s is superceded by %s, removing" %
                               (item, queue.queue[-1]))
                self.removeItem(item)

    def cycleForChange(self, *args, **kw):
        # Supercedent pipelines ignore circular dependencies and
        # individually enqueue each change that matches the trigger.
        # This is because they ignore shared queues and instead create
        # a virtual queue for each project-ref.
        return []

    def addChange(self, *args, **kw):
        ret = super(SupercedentPipelineManager, self).addChange(
            *args, **kw)
        if ret:
            self._pruneQueues()
        return ret

    def dequeueItem(self, item, quiet=False):
        super(SupercedentPipelineManager, self).dequeueItem(item, quiet)
        # A supercedent pipeline manager dynamically removes empty
        # queues
        if not item.queue.queue:
            self.state.removeQueue(item.queue)
